import sys
import time

import wx
import wx.lib.scrolledpanel
import wx.lib.filebrowsebutton as filebrowse

from . import buttons
from . import fonts
from .. import preferences
from ..editor import get_editors

import logging
log = logging.getLogger(__name__)


class PreferencesDialog(wx.Dialog):
    border = 3

    def __init__(self, parent, initial_page_name=None):
        wx.Dialog.__init__(self, parent, -1, "Preferences", size=(700, 400), pos=wx.DefaultPosition, style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        self.book = wx.Treebook(self, -1, style=wx.LB_LEFT)
        sizer.Add(self.book, 1, wx.ALL|wx.EXPAND, self.border)

        self.add_pages()
        if initial_page_name:
            self.show_page(initial_page_name)

        # self.Bind(wx.EVT_TREEBOOK_PAGE_CHANGED, self.OnPageChanged)
        # self.Bind(wx.EVT_TREEBOOK_PAGE_CHANGING, self.OnPageChanging)

        btnsizer = wx.StdDialogButtonSizer()
        self.ok_btn = wx.Button(self, wx.ID_OK)
        self.ok_btn.SetDefault()
        btnsizer.AddButton(self.ok_btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 0, wx.ALL|wx.EXPAND, self.border)

        self.Bind(wx.EVT_BUTTON, self.on_button)

        # Don't call self.Fit() otherwise the dialog buttons are zero height
        sizer.Fit(self)

    def add_pages(self):
        panel = ApplicationPreferencesPanel(self.book)
        self.book.AddPage(panel, panel.prefs.ui_name)

        editors = get_editors()
        for editor in editors:
            try:
                panel = EditorPreferencesPanel(self.book, editor, size=(500,500))
            except RuntimeError as e:
                # this editor has no preferences or preferences are same as its
                # superclass
                log.debug(f"Failed creating preference panel for {editor.ui_name}: {e}")
                pass
            else:
                self.book.AddPage(panel, editor.ui_name)

    def show_page(self, name):
        for index in range(self.book.GetPageCount()):
            if name == self.book.GetPageText(index):
                self.book.ChangeSelection(index)
                break

    def on_button(self, evt):
        if evt.GetId() == wx.ID_OK:
            self.commit_preferences()
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)
        evt.Skip()

    def commit_preferences(self):
        for i in range(self.book.GetPageCount()):
            panel = self.book.GetPage(i)
            panel.accept_preferences()


class InfoField:
    same_line = True
    display_label = True

    # wx.Sizer proportion of the main control (not the label).  See the
    # wx.Sizer docs, but basically 0 will fix vertical size to initial size, >
    # 0 will fill available space based on the total proportion in the sizer.
    vertical_proportion = 0

    default_width = 100

    popup_width = 300

    def __init__(self, panel, settings, desc, prefs, attrib_name):
        self.panel = panel
        self.field_name = desc
        self.prefs = prefs
        self.attrib_name = attrib_name
        self.create(settings)
        self.add_to_parent()

    def is_displayed(self, layer):
        return True

    def show(self, state=True):
        self.container.Show(state)

    def hide(self):
        self.show(False)

    def create(self, settings):
        self.container = wx.Window(self.panel)
        self.box = wx.BoxSizer(wx.VERTICAL)
        self.container.SetSizer(self.box)
        if self.display_label:
            self.label = wx.StaticText(self.container, label=self.field_name, style=wx.ST_ELLIPSIZE_END)
            bold_font = self.container.GetFont()
            bold_font.SetWeight(weight=wx.FONTWEIGHT_BOLD)
            self.label.SetFont(bold_font)
        self.create_all_controls(settings)
        if self.same_line and self.display_label:
            hbox = wx.BoxSizer(wx.HORIZONTAL)
            hbox.Add(self.label, 99, wx.ALIGN_CENTER)
            hbox.AddStretchSpacer(1)
            hbox.Add(self.ctrl, 0, wx.ALIGN_CENTER)
            for extra in self.extra_ctrls:
                hbox.Add(extra, 0, wx.ALIGN_CENTER)
            self.box.Add(hbox, 1, wx.EXPAND | wx.LEFT | wx.RIGHT, self.panel.SIDE_SPACING)
        else:
            if self.display_label:
                self.box.Add(self.label, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, self.panel.SIDE_SPACING)
                self.box.AddSpacer(self.panel.LABEL_SPACING)
            self.box.Add(self.ctrl, self.vertical_proportion, wx.EXPAND | wx.LEFT | wx.RIGHT, self.panel.SIDE_SPACING)
            for extra in self.extra_ctrls:
                self.box.Add(extra, 0, wx.ALIGN_CENTER)
        self.box.AddSpacer(self.panel.VALUE_SPACING)

    def create_all_controls(self, settings):
        self.ctrl = self.create_control(settings)
        self.extra_ctrls = self.create_extra_controls(settings)
        self.create_extra_event_handlers()

    def is_editable_control(self, ctrl):
        return ctrl == self.ctrl

    def create_extra_controls(self, settings):
        return []

    def create_extra_event_handlers(self):
        pass

    def add_to_parent(self):
        self.panel.sizer.Add(self.container, self.vertical_proportion, wx.EXPAND | wx.LEFT | wx.RIGHT | wx.ALIGN_TOP, 0)
        self.show(True)

    def get_value(self):
        return getattr(self.prefs, self.attrib_name)

    def fill_data(self):
        raise NotImplementedError

    def is_valid(self):
        return True

    def wants_focus(self):
        return False

    def set_focus(self):
        pass

    def process_command(self, cmd):
        # Override the normal refreshing of the InfoPanel when editing the
        # properties here because refreshing them messes up the text editing.
        self.panel.project.process_command(cmd, override_editable_properties_changed=False)


class LabelField(InfoField):
    alignment_style = wx.ALIGN_RIGHT

    def create_control(self, settings):
        c = wx.StaticText(self.container, style=self.alignment_style)
        return c


class TextEditField(InfoField):
    def create_control(self, settings):
        c = wx.TextCtrl(self.container)
        c.Bind(wx.EVT_TEXT, self.on_text_changed)
        c.SetEditable(True)
        return c

    def fill_data(self):
        try:
            text = self.get_value()
            self.ctrl.Enable(True)
        except IndexError:
            text = ""
            self.ctrl.Enable(False)
        self.ctrl.ChangeValue(text)
        self.is_valid()

    def is_valid(self):
        c = self.ctrl
        c.SetBackgroundColour("#FFFFFF")
        try:
            self.parse_from_ctrl()
            valid = True
        except Exception:
            c.SetBackgroundColour("#FF8080")
            valid = False
        self.ctrl.Refresh()
        return valid

    def parse_from_ctrl(self):
        return self.ctrl.GetValue()

    def on_text_changed(self, evt):
        if self.is_valid():
            value = self.parse_from_ctrl()
            setattr(self.prefs, self.attrib_name, value)

    def initial_text_input(self, text):
        self.ctrl.SetValue(text)
        self.ctrl.SetInsertionPointEnd()  # (self.ctrl.GetLastPosition())


class IntField(TextEditField):
    def get_value(self):
        value = getattr(self.prefs, self.attrib_name)
        text = str(value)
        return text

    def parse_from_ctrl(self):
        return int(self.ctrl.GetValue())


class BoolField(InfoField):
    def create_control(self, settings):
        c = wx.CheckBox(self.container)
        c.Bind(wx.EVT_CHECKBOX, self.on_toggle_changed)
        self.label.Bind(wx.EVT_LEFT_DOWN, self.on_label_clicked)
        return c

    def fill_data(self):
        state = self.get_value()
        self.ctrl.SetValue(state)
        self.is_valid()

    def is_valid(self):
        return True

    def parse_from_ctrl(self):
        return self.ctrl.GetValue()

    def on_toggle_changed(self, evt):
        if self.is_valid():
            value = self.parse_from_ctrl()
            setattr(self.prefs, self.attrib_name, value)

    def on_label_clicked(self, evt):
        self.ctrl.SetValue(not self.ctrl.GetValue())
        self.on_toggle_changed(evt)


class ChoiceField(InfoField):
    def fill_data(self):
        default_choice = self.get_value()
        self.ctrl.SetSelection(self.choices.index(default_choice))

    def create_control(self, settings):
        self.choices = settings
        c = wx.Choice(self.container, choices=[str(c) for c in self.choices])
        c.Bind(wx.EVT_CHOICE, self.drop_down_changed)
        return c

    def drop_down_changed(self, event):
        index = self.ctrl.GetSelection()
        setattr(self.prefs, self.attrib_name, self.choices[index])


class ColorPickerField(InfoField):
    same_line = True

    default_width = 100

    def fill_data(self):
        rgba = self.get_value()
        self.ctrl.SetColour(rgba)

    def create_control(self, settings):
        color = (0, 0, 0)
        c = buttons.ColorSelectButton(self.container, -1, "", color, size=(self.default_width, -1), style=wx.BORDER_NONE)
        c.Bind(buttons.EVT_COLORSELECT, self.on_color_changed)
        return c

    def on_color_changed(self, event):
        rgba = event.GetValue()
        color = wx.Colour(rgba)
        setattr(self.prefs, self.attrib_name, color)


class FontComboBox(wx.adv.OwnerDrawnComboBox):
    # Overridden from OwnerDrawnComboBox, called to draw each
    # item in the list
    def OnDrawItem(self, dc, rect, item, flags):
        if item == wx.NOT_FOUND:
            # painting the control, but there is no valid item selected yet
            return

        r = wx.Rect(*rect)  # make a copy
        r.Deflate(3, 5)

        face = fonts.get_font_name(item)
        font = wx.Font(12, wx.DEFAULT, wx.NORMAL, wx.NORMAL, False, face)
        dc.SetFont(font)

        if flags & wx.adv.ODCB_PAINTING_CONTROL:
            # for painting the control itself
            dc.DrawText(face, r.x + 5, (r.y + 5) + ((r.height / 2) - dc.GetCharHeight()) / 2)

        else:
            # for painting the items in the popup
            dc.DrawText(face,
                        r.x + 3,
                        (r.y + 5) + ((r.height / 2) - dc.GetCharHeight()) / 2
                        )

    # Overridden from OwnerDrawnComboBox, should return the height
    # needed to display an item in the popup, or -1 for default
    def OnMeasureItem(self, item):
        return 24

    # Overridden from OwnerDrawnComboBox.  Callback for item width, or
    # -1 for default/undetermined
    def OnMeasureItemWidth(self, item):
        return -1  # default - will be measured from text width


class FontField(InfoField):
    default_width = 200

    def fill_data(self):
        font = self.get_value()
        try:
            face = font.GetFaceName()
            index = fonts.get_font_index(face)
            self.ctrl.SetSelection(index)
            size = font.GetPointSize()
            index = fonts.standard_font_sizes.index(size)
            self.size_ctrl.SetSelection(index)
        except RuntimeError as e:
            log.error(f"Failed setting font to {font}: {e}")

    def create_control(self, settings):
        names = fonts.get_font_names()
        c = FontComboBox(self.container, -1, "", size=(self.default_width, -1), choices=names, style=wx.CB_READONLY)
        c.Bind(wx.EVT_COMBOBOX, self.on_face_changed)
        return c

    def create_extra_controls(self, settings):
        names = [str(s) for s in fonts.standard_font_sizes]
        c = wx.ComboBox(self.container, -1, str(fonts.default_font_size), size=(self.default_width/2, -1), choices=names, style=wx.CB_READONLY)
        c.Bind(wx.EVT_COMBOBOX, self.on_size_changed)
        self.size_ctrl = c
        return [c]

    def on_face_changed(self, event):
        self.set_font()

    def on_size_changed(self, event):
        self.set_font()

    def set_font(self):
        index = self.ctrl.GetSelection()
        face = fonts.get_font_name(index)
        index = self.size_ctrl.GetSelection()
        size = fonts.standard_font_sizes[index]
        font = wx.Font(size, wx.DEFAULT, wx.NORMAL, wx.NORMAL, False, face)
        setattr(self.prefs, self.attrib_name, font)


class IntRangeField(InfoField):
    default_width = 200

    def fill_data(self):
        value = self.get_value()
        self.ctrl.SetValue(value)

    def get_params(self, settings):
        try:
            lo, hi = settings.split("-")
            lo = int(lo.strip())
            hi = int(hi.strip())
        except:
            lo, hi = 0, 100
        return lo, hi

    def create_control(self, settings):
        minval, maxval = self.get_params(settings)
        c = wx.Slider(self.container, -1, minval, minval, maxval, size=(self.default_width, -1), style=wx.SL_HORIZONTAL | wx.SL_AUTOTICKS | wx.SL_LABELS)
        c.Bind(wx.EVT_SLIDER, self.slider_changed)
        return c

    def slider_changed(self, event):
        setattr(self.prefs, self.attrib_name, self.ctrl.GetValue())


class DirectoryField(InfoField):
    same_line = True

    default_width = 400

    def fill_data(self):
        path = self.get_value()
        self.ctrl.SetValue(path, False)

    def create_control(self, settings):
        c = filebrowse.DirBrowseButton(self.container, -1, size=(self.default_width, -1), labelText="", changeCallback = self.on_directory_changed)
        return c

    def on_directory_changed(self, event):
        path = event.GetString()
        setattr(self.prefs, self.attrib_name, path)


known_fields = {
    "int": IntField,
    "intrange": IntRangeField,
    "bool": BoolField,
    "wx.Colour": ColorPickerField,
    "Color": ColorPickerField,
    "wx.Font": FontField,
    "Font": FontField,
    "directory": DirectoryField,
}


def register_preference_field(name, cls):
    global known_fields

    known_fields[name] = cls


def find_field(field_type):
    settings = None
    if ":" in field_type:
        field_type, settings = field_type.split(":", 1)
    try:
        field_cls = known_fields[field_type]
    except TypeError:
        field_cls = ChoiceField
        settings = field_type[:]
    except KeyError:
        log.warning(f"Unknown preference type {field_type}")
        raise
    return field_cls, settings

def calc_field(parent, prefs, attrib_name, field_type, desc):
    field_cls, settings = find_field(field_type)
    log.debug(f"calc_field: {field_type}: {field_cls.__class__.__name__}, {settings}")
    field = field_cls(parent, settings, desc, prefs, attrib_name)
    field.fill_data()
    return field


PANELTYPE = wx.lib.scrolledpanel.ScrolledPanel
class PreferencesPanel(PANELTYPE):
    """
    A panel for displaying and manipulating the properties of a layer.
    """
    LABEL_SPACING = 10
    VALUE_SPACING = 3
    SIDE_SPACING = 5

    window_name = "InfoPanel"

    def __init__(self, parent, prefs, size=(-1,-1)):
        PANELTYPE.__init__(self, parent, name=self.window_name, size=size)
        self.prefs = prefs

        # Mac/Win needs this, otherwise background color is black
        attr = self.GetDefaultAttributes()
        self.SetBackgroundColour(attr.colBg)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.sizer)

        self.display_panel_for_prefs()

    def display_panel_for_prefs(self):
        self.sizer.AddSpacer(self.LABEL_SPACING)

        prefs = self.prefs
        self.fields = []
        for field_info in prefs.display_order:
            attrib_cls = None
            desc = None
            if isinstance(field_info, str):
                attrib_name = field_info
            else:
                try:
                    attrib_name, attrib_cls_name, desc = field_info
                except ValueError:
                    try:
                        attrib_name, attrib_cls_name = field_info
                    except ValueError:
                        attrib_name = field_info[0]
            if desc is None:
                desc = attrib_name.replace("_", " ").capitalize()
            if attrib_cls_name is None:
                attrib = getattr(prefs, attrib_name)
                attrib_cls_name = attrib.__class__.__name__
            log.debug(f"{self.prefs._module_path}: {attrib_name}={attrib_cls_name}")

            try:
                field = calc_field(self, prefs, attrib_name, attrib_cls_name, desc)
                self.fields.append(field)
            except KeyError:
                pass

        self.sizer.Layout()

    def get_source_preferences(self):
        pass

    def accept_preferences(self):
        source = self.get_source_preferences()
        if source != self.prefs:
            log.debug(f"preferences updated! {source._module_path}")
            source.copy_from(self.prefs)
            self.prefs.persist_user_settings()


class ApplicationPreferencesPanel(PreferencesPanel):
    def __init__(self, parent, size=(-1,-1)):
        prefs = wx.GetApp().get_preferences().clone()
        PreferencesPanel.__init__(self, parent, prefs, size=size)

    def get_source_preferences(self):
        return wx.GetApp().get_preferences()


class EditorPreferencesPanel(PreferencesPanel):
    def __init__(self, parent, editor, size=(-1,-1)):
        self.editor = editor
        prefs = editor.get_preferences().clone()
        if not prefs.display_order:
            raise RuntimeError("No displayed preferences")
        s = editor.__mro__[1]
        if s.preferences_module == editor.preferences_module:
            raise RuntimeError("Unchanged preferences from superclass")
        PreferencesPanel.__init__(self, parent, prefs, size=size)

    def get_source_preferences(self):
        return self.editor.get_preferences()



import jsonpickle
import jsonpickle.ext.numpy as jsonpickle_numpy
jsonpickle_numpy.register_handlers()

import logging
log = logging.getLogger(__name__)
#log.setLevel(logging.DEBUG)

def collapse_json(text, indent=8, special_keys={}, base_indent=4):
    """Compacts a string of json data by collapsing whitespace after the
    specified indent level with optional keys that can be expanded beyond the
    indent level.
    
    NOTE: will not produce correct results when indent level is not a multiple
    of the json indent level
    """
    initial = " " * indent
    out = []  # final json output
    sublevel = []  # accumulation list for sublevel entries
    pending = None  # holder for consecutive entries at exact indent level
    special = False # are we processing a block that should be expanded
    end_level = 0  # trigger value to determine end of expanded block
    special_lines = []
    special_indent = 0
    remove_next_indent = True  # flag for special case to add indent to line following expanded block
    for line in text.splitlines():
        first_non_blank = len(line) - len(line.lstrip())
        if special:
            special_lines.append(line)
            if first_non_blank == end_level:
                subtext = "\n".join(special_lines)
                log.debug("formatting subblock with %d" % special_indent)
                log.debug(subtext)
                subtext = collapse_json(subtext, special_indent)
                log.debug("formatted subblock:")
                log.debug(subtext)
                log.debug("Resetting to normal.")
                special = False
                out.append(subtext)
                remove_next_indent = False
            continue
        if line.startswith(initial):

            # special expanded keys must be dictionary entries in the json, so
            # they will end with ':'
            keys = line.strip().split(":")
            if not special and len(keys) > 1:
                key = keys[0][1:-1]  # remove quotes
                if key in special_keys:
                    extra_levels = special_keys[key]
                    special_indent = (extra_levels * base_indent) + first_non_blank
                    log.debug("Found %s: %d*%d + %d = %d" % (key, extra_levels, base_indent, indent, special_indent))
                    special = True
                    end_level = first_non_blank
                    special_lines = [line]
                    if sublevel:
                        out.append("".join(sublevel))
                        sublevel = []
                    if pending:
                        out.append(pending)
                        pending = None
                    continue

            # lines following a specially indented block wouldn't ordinarily
            # have an indent, so force it with this flag
            if remove_next_indent:
                item = line.strip()
            else:
                item = line.rstrip()

            if line[indent] == " ":
                # found a line indented further than the indent level, so add
                # it to the sublevel list
                if pending:
                    # the first item in the sublevel will be the pending item
                    # that was the previous line in the json
                    sublevel.append(pending)
                    pending = None
                sublevel.append(item)
                if item.endswith(","):
                    sublevel.append(" ")
            elif sublevel:
                # found a line at the exact indent level *and* we have sublevel
                # items. This means the sublevel items have come to an end
                sublevel.append(item)
                out.append("".join(sublevel))
                sublevel = []
            else:
                # found a line at the exact indent level but no items indented
                # further, so possibly start a new sub-level
                if pending:
                    # if there is already a pending item, it means that
                    # consecutive entries in the json had the exact same
                    # indentation and that last pending item was not the start
                    # of a new sublevel.
                    out.append(pending)
                pending = line.rstrip()
        else:
            if pending:
                # it's possible that an item will be pending but not added to
                # the output yet, so make sure it's not forgotten.
                out.append(pending)
                pending = None
            if sublevel:
                out.append("".join(sublevel))
            out.append(line)
        remove_next_indent = True
    if sublevel:
        out.append("".join(sublevel))
    return "\n".join(out)


def dict_to_list(d):
    """Return a sorted list of lists that represent the dictionary.

    Only strings are allowed as keys in a json dictionary, so this transform
    the dictionary into a list of lists that can be serialized in as small
    amount of text as possible. Tuples would be more efficient in python terms,
    but json can't serialize tuples directly either.
    """
    return sorted([list(i) for i in list(d.items())])

def unserialize(name, text):
    try:
        if hasattr(text, 'decode'):
            text = text.decode('utf-8')
        if text.startswith("#"):
            header, text = text.split("\n", 1)
        unserialized = jsonpickle.loads(text)
    except ValueError as e:
        log.error(f"JSON parsing error for extra metadata: {str(e)}")
        import traceback
        traceback.print_exc()
        unserialized = {}
    except AttributeError as e:
        log.error(f"JSON library error: {str(e)}")
        import traceback
        traceback.print_exc()
        unserialized = {}
    return unserialized

if __name__ == "__main__":
    import json
    logging.basicConfig(level=logging.WARNING)


    s = {"zero": ["first", {"second": 2, "third": 3, "fourth": 4, "items": [[1,2,3,4], [5,6,7,8], 9, 10, [11, [12, [13, [14, 15]]]]], "items2": [[1,2,3,4], [5,6,7,8], 9, 10, [11, [12, [13, [14, 15]]]]]}],"zeroprime": [10,[12,[14, 16]]]}

    text = json.dumps(s, indent=4)
    print("original")
    print(text)

    def process(level):
        processed = collapse_json(text, indent=level, special_keys = {"viewers": 1, "linked bases": 1, "items2": 1})
        print(level)
        print(processed)
        rebuilt = json.loads(processed)
        assert rebuilt == s

    for level in range(0, 21, 4):
        process(level)



# pyparsing code based on original code arith.py by Paul McGuire, modified by
# Glenn Linderman, licensed under the pyparsing license arith.py from:
#
# http://pyparsing.wikispaces.com/file/view/arith.py/241810293/arith.py
import numpy as np

from pyparsing import Word, nums, hexnums, alphas, alphanums, Combine, oneOf, Optional, \
    opAssoc, operatorPrecedence, ParseException, ParserElement, Literal, Regex, pyparsing_common

ParserElement.enablePackrat()


class EvalConstant():
    "Class to evaluate a parsed constant or variable"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_):
        v = self.value
        if v in vars_:
            return vars_[v]
        else:
            if v.startswith("$"):
                return int(v[1:], 16)
            elif v.startswith("0x"):
                return int(v[2:], 16)
            else:
                return int(v)


class EvalSignOp():
    "Class to evaluate expressions with a leading + or - sign"

    def __init__(self, tokens):
        self.sign, self.value = tokens[0]

    def eval(self, vars_):
        mult = {'+':1, '-':-1}[self.sign]
        return mult * self.value.eval(vars_)


def operatorOperands(tokenlist):
    "generator to extract operators and operands in pairs"
    it = iter(tokenlist)
    while 1:
        try:
            o1 = next(it)
            o2 = next(it)
            yield (o1, o2)
        except StopIteration:
            break


class EvalMultOp():
    "Class to evaluate multiplication and division expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        prod = self.value[0].eval(vars_)
        for op,val in operatorOperands(self.value[1:]):
            if op == '*':
                prod *= val.eval(vars_)
            if op == '/':
                prod /= val.eval(vars_)
            if op == '//':
                prod //= val.eval(vars_)
            if op == '%':
                prod %= val.eval(vars_)
        return prod


class EvalAddOp():
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        sum = self.value[0].eval(vars_)
        for op,val in operatorOperands(self.value[1:]):
            if op == '+':
                sum += val.eval(vars_)
            if op == '-':
                sum -= val.eval(vars_)
        return sum


class EvalBitwiseAndOp():
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        val1 = self.value[0].eval(vars_)
        for op,val in operatorOperands(self.value[1:]):
            val2 = val.eval(vars_)
            val1 = val1 & val2
        return val1


class EvalBitwiseOrOp():
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        val1 = self.value[0].eval(vars_)
        for op,val in operatorOperands(self.value[1:]):
            val2 = val.eval(vars_)
            val1 = val1 | val2
        return val1


class EvalLogicalAndOp():
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        val1 = self.value[0].eval(vars_)
        for op,val in operatorOperands(self.value[1:]):
            val2 = val.eval(vars_)
            val1 = np.logical_and(val1, val2)
        return val1


class EvalLogicalOrOp():
    "Class to evaluate addition and subtraction expressions"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        val1 = self.value[0].eval(vars_)
        for op,val in operatorOperands(self.value[1:]):
            val2 = val.eval(vars_)
            val1 = np.logical_or(val1, val2)
        return val1


class EvalComparisonOp():
    "Class to evaluate comparison expressions"
    opMap = {
        "<" : lambda a,b : a < b,
        "<=" : lambda a,b : a <= b,
        ">" : lambda a,b : a > b,
        ">=" : lambda a,b : a >= b,
        "==" : lambda a,b : a == b,
        "!=" : lambda a,b : a != b,
        "<>" : lambda a,b : a != b,
        }

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_ ):
        val1 = self.value[0].eval(vars_)
        # don't use "type(val1) is np.ndarray" because subclasses will fail!
        if isinstance(val1, np.ndarray):
            for op,val in operatorOperands(self.value[1:]):
                fn = self.opMap[op]
                val2 = val.eval(vars_)
                val1 = fn(val1, val2)
            return val1
        else:
            for op,val in operatorOperands(self.value[1:]):
                fn = self.opMap[op]
                val2 = val.eval(vars_)
                if not fn(val1,val2):
                    break
                val1 = val2
            else:
                return True
            return False


class NumpyIntExpression():
    integer = Word(nums)
    hexint = Combine(oneOf('0x $') + Word(hexnums))

    variable = Word(alphas)
    operand = hexint | integer | variable

    signop = oneOf('+ -')
    multop = oneOf('* / // %')
    plusop = oneOf('+ -')
    bitwiseandop = Literal('&')
    bitwiseorop = Literal('|')
    logicalandop = oneOf('and &&')
    logicalorop = oneOf('or ||')
    comparisonop = oneOf("< <= > >= == != <>")

    # use parse actions to attach EvalXXX constructors to sub-expressions
    operand.setParseAction(EvalConstant)
    arith_expr = operatorPrecedence(operand,
        [(signop, 1, opAssoc.RIGHT, EvalSignOp),
         (multop, 2, opAssoc.LEFT, EvalMultOp),
         (plusop, 2, opAssoc.LEFT, EvalAddOp),
         (bitwiseandop, 2, opAssoc.LEFT, EvalBitwiseAndOp),
         (bitwiseorop, 2, opAssoc.LEFT, EvalBitwiseOrOp),
         (comparisonop, 2, opAssoc.LEFT, EvalComparisonOp),
         (logicalandop, 2, opAssoc.LEFT, EvalLogicalAndOp),
         (logicalorop, 2, opAssoc.LEFT, EvalLogicalOrOp),
         ])

    def __init__(self, vars_={}):
        self.vars_ = vars_

    def setvars(self, vars_):
        self.vars_ = vars_

    def setvar( var, val ):
        self.vars_[ var ] = val

    def eval( self, strExpr ):
        ret = self.arith_expr.parseString( strExpr, parseAll=True)[0]
        result = ret.eval( self.vars_ )
        return result


class EvalFloatConstant():
    "Class to evaluate a parsed constant or variable"

    def __init__(self, tokens):
        self.value = tokens[0]

    def eval(self, vars_):
        v = self.value
        if v in vars_:
            return vars_[v]
        else:
            return float(v)


class NumpyFloatExpression():
    variable = Word(alphas+"_", alphanums+"_")
    leading_dot_float = Regex(r'\.\d+')
    operand = leading_dot_float | pyparsing_common.number | variable

    signop = oneOf('+ -')
    multop = oneOf('* / // %')
    plusop = oneOf('+ -')
    andop = oneOf('and &')
    orop = oneOf('or |')
    comparisonop = oneOf("< <= > >= == != <>")

    # use parse actions to attach EvalXXX constructors to sub-expressions
    operand.setParseAction(EvalFloatConstant)
    arith_expr = operatorPrecedence(operand,
        [(signop, 1, opAssoc.RIGHT, EvalSignOp),
         (multop, 2, opAssoc.LEFT, EvalMultOp),
         (plusop, 2, opAssoc.LEFT, EvalAddOp),
         (comparisonop, 2, opAssoc.LEFT, EvalComparisonOp),
         (andop, 2, opAssoc.LEFT, EvalLogicalAndOp),
         (orop, 2, opAssoc.LEFT, EvalLogicalOrOp),
         ])

    def __init__(self, vars_={}):
        self.vars_ = vars_

    def setvars(self, vars_):
        self.vars_ = vars_

    def setvar( var, val ):
        self.vars_[ var ] = val

    def eval( self, strExpr ):
        ret = self.arith_expr.parseString( strExpr, parseAll=True)[0]
        result = ret.eval( self.vars_ )
        return result


if __name__=='__main__':
    a = np.arange(256*256)
    b = np.arange(256*256)
    f = np.linspace(0, 1, 101, dtype=np.float32)
    v = {"a": a, "b": b, "f": f}
    arith = NumpyIntExpression(v)
    tests = [
        ("a > 3", a > 3),
        ("b > 8", b > 8),
        ("(a > 3) && (a > 5)", a > 5),
        ("a > 3 && a > 5", a > 5),
        ("(a > 3) && (a < 100)", np.logical_and((a > 3), (a < 100))),
        ("a > 3 && a < 100", np.logical_and((a > 3), (a < 100))),
        ("(a & 7)", a & 7),
        ("((a & 7) > 3)", (a & 7) > 3),
        ("((a & 7) > 3) && (a > 5)", np.logical_and(((a & 7) > 3), (a > 5))),
        ("((a & 7) > 3) || (a > 25)", np.logical_or(((a & 7) > 3), (a > 25))),
        ("(a > 1000) || (a < 20)", np.logical_or((a > 1000), (a < 20))),
        ("a > 1000 || a < 20", np.logical_or((a > 1000), (a < 20))),
        ]
    for test, expected in tests:
        result = arith.eval(test)
        correct = (expected == result).all()
        if correct:
            print((test, "correct!"))
        else:
            print((test, "FAILED", expected, result))

    arith = NumpyFloatExpression(v)
    tests = [
        ("f > 3", f > 3),
        ("f > 0.3", f > .3),
        ("f > .3", f > .3),
        ]
    for test, expected in tests:
        result = arith.eval(test)
        correct = (expected == result).all()
        if correct:
            print((test, "correct!"))
        else:
            print((test, "FAILED", expected, result))



import sys
import time
import importlib
import inspect

import jsonpickle

import wx

# Mapping of strings to valid wxFont families
font_families = {
    'default': wx.FONTFAMILY_DEFAULT,
    'decorative': wx.FONTFAMILY_DECORATIVE,
    'roman': wx.FONTFAMILY_ROMAN,
    'script': wx.FONTFAMILY_SCRIPT,
    'swiss': wx.FONTFAMILY_SWISS,
    'modern': wx.FONTFAMILY_MODERN
}

# Mapping of strings to wxFont styles
font_styles = {
    'slant': wx.FONTSTYLE_SLANT,
    'italic': wx.FONTSTYLE_ITALIC
}

# Mapping of strings wxFont weights
font_weights = {
    'light': wx.FONTWEIGHT_LIGHT,
    'bold': wx.FONTWEIGHT_BOLD
}

# Strings to ignore in text representations of fonts
font_noise = ['pt', 'point', 'family']

standard_font_sizes = [4, 6, 8, 9, 10, 11, 12, 14, 16, 18, 20, 22, 24, 28, 32, 36, 40, 48, 56, 64, 72, 144]


if sys.platform == "darwin":
    default_font = "10 point Monaco"
elif sys.platform == "win32":
    default_font = "10 point Lucida Console"
else:
    default_font = "10 point monospace"

default_font_size = 12
default_font_index = standard_font_sizes.index(default_font_size)


# Font conversion utilities from Traitsui

def font_to_str(font):
    """ Converts a wx.Font into a string description of itself.
    """
    weight = {wx.FONTWEIGHT_LIGHT: ' Light',
              wx.FONTWEIGHT_BOLD: ' Bold'}.get(font.GetWeight(), '')
    style = {wx.FONTSTYLE_SLANT: ' Slant',
             wx.FONTSTYLE_ITALIC: ' Italic'}.get(font.GetStyle(), '')
    underline = ''
    if font.GetUnderlined():
        underline = ' underline'
    return '%s point %s%s%s%s' % (
           font.GetPointSize(), font.GetFaceName(), style, weight, underline)


def str_to_font(value):
    """ Create a wx.Font object from a string description.
    """
    point_size = None
    family = wx.FONTFAMILY_DEFAULT
    style = wx.FONTSTYLE_NORMAL
    weight = wx.FONTWEIGHT_NORMAL
    underline = 0
    facename = []
    for word in value.split():
        lword = word.lower()
        if lword in font_families:
            family = font_families[lword]
        elif lword in font_styles:
            style = font_styles[lword]
        elif lword in font_weights:
            weight = font_weights[lword]
        elif lword == 'underline':
            underline = 1
        elif lword not in font_noise:
            if point_size is None:
                try:
                    point_size = int(lword)
                    continue
                except:
                    pass
            facename.append(word)
    return wx.Font(point_size or 10, family, style, weight, underline, ' '.join(facename))


known_fonts = None  # Will be initialized in call to get_font_names


def get_font_names():
    global known_fonts

    if known_fonts is None:
        # NOTE: The list of face names from wx are in unicode
        fonts = wx.FontEnumerator()
        fonts.EnumerateFacenames()
        known_fonts = fonts.GetFacenames()
        known_fonts.sort()
        known_fonts[0:0] = ["default"]
    return known_fonts


def get_font_name(index):
    fonts = get_font_names()
    return fonts[index]


def get_font_index(face):
    fonts = get_font_names()
    try:
        return fonts.index(face)
    except ValueError:
        return 0


class wxFontHandler(jsonpickle.handlers.BaseHandler):
    def flatten(self, obj, data):
        data["font"] = font_to_str(obj)
        return data

    def restore(self, obj):
        font_str = obj["font"]
        return str_to_font(font_str)

jsonpickle.handlers.registry.register(wx.Font, wxFontHandler)



import sys
import argparse

import logging
log = logging.getLogger(__name__)

import wx


def setup_frozen_logging():
    # set up early py2exe logging redirection, saving any messages until the log
    # file directory can be determined after the application is initialized.
    frozen = getattr(sys, 'frozen', False)
    if frozen in ('dll', 'windows_exe', 'console_exe'):
        class Blackhole(object):
            softspace = 0
            saved_text = []

            def write(self, text):
                self.saved_text.append(text)

            def flush(self):
                pass
        sys.stdout = Blackhole()
        sys.stderr = sys.stdout


def run(app_cls, image_paths=[], template_paths=[], help_paths=[], debug_log=False):
    from . import filesystem

    filesystem.image_paths.extend(image_paths)
    filesystem.template_paths.extend(template_paths)
    filesystem.help_paths.extend(help_paths)

    # Create a debugging log
    if debug_log:
        filename = app.get_log_file_name("debug")
        handler = logging.FileHandler(filename)
        logger = logging.getLogger('')
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

    # Turn off omnivore log debug messages by default
    log = logging.getLogger("sawx")
    log.setLevel(logging.INFO)

    app = app_cls()

    app.process_command_line_args(sys.argv[1:])

    app.MainLoop()

    from .utils.jobs import get_global_job_manager
    job_manager = get_global_job_manager()
    if job_manager is not None:
        job_manager.shutdown()



#-----------------------------------------------------------------------------
# Name:        dropscroller.py
# Purpose:     auto scrolling for a list that's being used as a drop target
#
# Author:      Rob McMullen
#
# Created:     2007
# RCS-ID:      $Id: $
# Copyright:   (c) 2007-2016 Rob McMullen
# License:     wxPython
#-----------------------------------------------------------------------------
"""
Automatic scrolling mixin for a list control, including an indicator
showing where the items will be dropped.

It would be nice to have somethin similar for a tree control as well,
but I haven't tackled that yet.
"""
import sys, pickle

import wx


class ListDropScrollerMixin(object):
    """Automatic scrolling for ListCtrls for use when using drag and drop.

    This mixin is used to automatically scroll a list control when
    approaching the top or bottom edge of a list.  Currently, this
    only works for lists in report mode.

    Add this as a mixin in your list, and then call processListScroll
    in your DropTarget's OnDragOver method.  When the drop ends, call
    finishListScroll to clean up the resources (i.e. the wx.Timer)
    that the dropscroller uses and make sure that the insertion
    indicator is erased.

    The parameter interval is the delay time in milliseconds between
    list scroll steps.

    If indicator_width is negative, then the indicator will be the
    width of the list.  If positive, the width will be that number of
    pixels, and zero means to display no indicator.
    """

    def __init__(self, interval=200, width=-1):
        """Don't forget to call this mixin's init method in your List.

        Interval is in milliseconds.
        """
        self._auto_scroll_timer = None
        self._auto_scroll_interval = interval
        self._auto_scroll = 0
        self._auto_scroll_save_y = -1
        self._auto_scroll_save_width = width
        self._auto_scroll_last_state = 0
        self._auto_scroll_last_index = -1
        self._auto_scroll_indicator_line = True
        self.Bind(wx.EVT_TIMER, self.OnAutoScrollTimer)

    def clearAllSelected(self):
        """clear all selected items"""
        list_count = self.GetItemCount()
        for index in range(list_count):
            self.SetItemState(index, 0, wx.LIST_STATE_SELECTED)

    def _startAutoScrollTimer(self, direction = 0):
        """Set the direction of the next scroll, and start the
        interval timer if it's not already running.
        """
        if self._auto_scroll_timer == None:
            self._auto_scroll_timer = wx.Timer(self, wx.TIMER_ONE_SHOT)
            self._auto_scroll_timer.Start(self._auto_scroll_interval)
        self._auto_scroll = direction

    def _stopAutoScrollTimer(self):
        """Clean up the timer resources.
        """
        self._auto_scroll_timer = None
        self._auto_scroll = 0

    def _getAutoScrollDirection(self, index):
        """Determine the scroll step direction that the list should
        move, based on the index reported by HitTest.
        """
        first_displayed = self.GetTopItem()

        if first_displayed == index:
            # If the mouse is over the first index...
            if index > 0:
                # scroll the list up unless...
                return -1
            else:
                # we're already at the top.
                return 0
        elif index >= first_displayed + self.GetCountPerPage() - 1:
            # If the mouse is over the last visible item, but we're
            # not at the last physical item, scroll down.
            return 1
        # we're somewhere in the middle of the list.  Don't scroll
        return 0

    def getDropIndex(self, x, y, index=None, flags=None, insert=True):
        """Find the index to insert the new item, which could be
        before or after the index passed in.
        
        @return: if insert is true, return value is the index of the insert
        point (i.e.  the new data should be inserted at that point, pushing
        the existing data further down the list).  If insert is false, the
        return value is the item on which the drop happened, or -1 indicating
        an invalid drop.
        """
        if index is None:
            index, flags = self.HitTest((x, y))

        # Not clicked on an item
        if index == wx.NOT_FOUND:

            # If it's an empty list or below the last item
            if (flags & (wx.LIST_HITTEST_NOWHERE|wx.LIST_HITTEST_ABOVE|wx.LIST_HITTEST_BELOW)):

                # Append to the end of the list or return an invalid index
                if insert:
                    index = self.GetItemCount()
                else:
                    index = self.GetItemCount() - 1
                #print "getDropIndex: append to end of list: index=%d" % index
            elif (self.GetItemCount() > 0):
                if y <= self.GetItemRect(0).y: # clicked just above first item
                    index = 0 # append to top of list
                    #print "getDropIndex: before first item: index=%d, y=%d, rect.y=%d" % (index, y, self.GetItemRect(0).y)
                elif insert:
                    index = self.GetItemCount() # append to end of list
                    #print "getDropIndex: after last item: index=%d" % index
                else:
                    index = self.GetItemCount() - 1

        # Otherwise, we've clicked on an item.  If we're in insert mode, check
        # to see if the cursor is between items
        elif insert:
            # Get bounding rectangle for the item the user is dropping over.
            rect = self.GetItemRect(index)
            #print "getDropIndex: landed on %d, y=%d, rect=%s" % (index, y, rect)

            # NOTE: On all platforms, the y coordinate used by HitTest
            # is relative to the scrolled window.  There are platform
            # differences, however, because on GTK the top of the
            # vertical scrollbar stops below the header, while on MSW
            # the top of the vertical scrollbar is equal to the top of
            # the header.  The result is the y used in HitTest and the
            # y returned by GetItemRect are offset by a certain amount
            # on GTK.  The HitTest's y=0 in GTK corresponds to the top
            # of the first item, while y=0 on MSW is in the header.

            # From Robin Dunn: use GetMainWindow on the list to find
            # the actual window on which to draw
            if self != self.GetMainWindow():
                y += self.GetMainWindow().GetPosition().Get()[1]

            # If the user is dropping into the lower half of the rect,
            # we want to insert _after_ this item.
            if y >= (rect.y + rect.height/2):
                index = index + 1

        return index

    def processListScroll(self, x, y, line=True):
        """Main handler: call this with the x and y coordinates of the
        mouse cursor as determined from the OnDragOver callback.

        This method will determine which direction the list should be
        scrolled, and start the interval timer if necessary.
        """
        if self.GetItemCount() == 0:
            # don't show any lines if we don't have any items in the list
            return

        index, flags = self.HitTest((x, y))

        direction = self._getAutoScrollDirection(index)
        if direction == 0:
            self._stopAutoScrollTimer()
        else:
            self._startAutoScrollTimer(direction)
        self._auto_scroll_indicator_line = line

        drop_index = self.getDropIndex(x, y, index=index, flags=flags)
        if line:
            self._processLineIndicator(x, y, drop_index)
        else:
            self._processHighlightIndicator(drop_index)

    def _processLineIndicator(self, x, y, drop_index):
        count = self.GetItemCount()
        if drop_index >= count:
            index = min(count, drop_index)
            rect = self.GetItemRect(index - 1)
            y = rect.y + rect.height + 1
        else:
            rect = self.GetItemRect(drop_index)
            y = rect.y

        # From Robin Dunn: on GTK & MAC the list is implemented as
        # a subwindow, so have to use GetMainWindow on the list to
        # find the actual window on which to draw
        if self != self.GetMainWindow():
            y -= self.GetMainWindow().GetPosition().Get()[1]

        if self._auto_scroll_save_y == -1 or self._auto_scroll_save_y != y:
            #print "main window=%s, self=%s, pos=%s" % (self, self.GetMainWindow(), self.GetMainWindow().GetPosition().Get())
            if self._auto_scroll_save_width < 0:
                self._auto_scroll_save_width = rect.width
            dc = self._getIndicatorDC()
            self._eraseIndicator(dc)
            dc.DrawLine(0, y, self._auto_scroll_save_width, y)
            self._auto_scroll_save_y = y

    def _processHighlightIndicator(self, index):
        count = self.GetItemCount()
        if index >= count:
            index = count - 1
        if self._auto_scroll_last_index != index:
            selected = self.GetItemState(index, wx.LIST_STATE_SELECTED)
            if not selected:
                self.SetItemState(index, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)
            if self._auto_scroll_last_index != -1:
                #self.SetItemState(self._auto_scroll_last_index, self._auto_scroll_last_state, wx.LIST_STATE_SELECTED)
                self.SetItemState(self._auto_scroll_last_index, 0, wx.LIST_STATE_SELECTED)
            self._auto_scroll_last_state = selected
            self._auto_scroll_last_index = index

    def finishListScroll(self):
        """Clean up timer resource and erase indicator.
        """
        self._stopAutoScrollTimer()
        self._eraseIndicator()
        self._auto_scroll_last_index = -1
        self._auto_scroll_last_state = 0

    def OnAutoScrollTimer(self, evt):
        """Timer event handler to scroll the list in the requested
        direction.
        """
        #print "_auto_scroll = %d, timer = %s" % (self._auto_scroll, self._auto_scroll_timer is not None)
        count = self.GetItemCount()
        if self._auto_scroll == 0:
            # clean up timer resource
            self._auto_scroll_timer = None
        elif count > 0:
            if self._auto_scroll_indicator_line:
                dc = self._getIndicatorDC()
                self._eraseIndicator(dc)
            if self._auto_scroll < 0:
                index = max(0, self.GetTopItem() + self._auto_scroll)
            else:
                index = min(self.GetTopItem() + self.GetCountPerPage(), count - 1)
            self.EnsureVisible(index)
            self._auto_scroll_timer.Start()
        evt.Skip()

    def _getIndicatorDC(self):
        dc = wx.ClientDC(self.GetMainWindow())
        dc.SetPen(wx.Pen(wx.WHITE, 3))
        dc.SetBrush(wx.TRANSPARENT_BRUSH)
        dc.SetLogicalFunction(wx.XOR)
        return dc

    def _eraseIndicator(self, dc=None):
        if self._auto_scroll_indicator_line:
            if dc is None:
                dc = self._getIndicatorDC()
            if self._auto_scroll_save_y >= 0:
                # erase the old line
                dc.DrawLine(0, self._auto_scroll_save_y,
                            self._auto_scroll_save_width, self._auto_scroll_save_y)
        self._auto_scroll_save_y = -1


class PickledDataObject(wx.CustomDataObject):
    """Sample custom data object storing indexes of the selected items"""

    def __init__(self):
        wx.CustomDataObject.__init__(self, "Pickled")


class PickledDropTarget(wx.PyDropTarget):
    """Custom drop target modified from the wxPython demo."""
    debug = False

    def __init__(self, window):
        wx.PyDropTarget.__init__(self)
        self.dv = window

        # specify the type of data we will accept
        self.data = PickledDataObject()
        self.SetDataObject(self.data)

    def cleanup(self):
        self.dv.finishListScroll()

    # some virtual methods that track the progress of the drag
    def OnEnter(self, x, y, d):
        if self.debug: print(("OnEnter: %d, %d, %d\n" % (x, y, d)))
        return d

    def OnLeave(self):
        if self.debug: print("OnLeave\n")
        self.cleanup()

    def OnDrop(self, x, y):
        if self.debug: print(("OnDrop: %d %d\n" % (x, y)))
        self.cleanup()
        return True

    def OnDragOver(self, x, y, d):
        top = self.dv.GetTopItem()
        if self.debug: print(("OnDragOver: %d, %d, %d, top=%s" % (x, y, d, top)))

        self.dv.processListScroll(x, y)

        # The value returned here tells the source what kind of visual
        # feedback to give.  For example, if wxDragCopy is returned then
        # only the copy cursor will be shown, even if the source allows
        # moves.  You can use the passed in (x,y) to determine what kind
        # of feedback to give.  In this case we return the suggested value
        # which is based on whether the Ctrl key is pressed.
        return d

    # Called when OnDrop returns True.  We need to get the data and
    # do something with it.
    def OnData(self, x, y, d):
        if self.debug: print(("OnData: %d, %d, %d\n" % (x, y, d)))

        self.cleanup()
        # copy the data from the drag source to our data object
        if self.GetData():
            # convert it back to a list of lines and give it to the viewer
            items = pickle.loads(self.data.GetData())
            self.dv.add_dropped_items(x, y, items)

        # what is returned signals the source what to do
        # with the original data (move, copy, etc.)  In this
        # case we just return the suggested value given to us.
        return d


from wx.lib.mixins import listctrl


class ReorderableList(wx.ListCtrl, listctrl.ListCtrlAutoWidthMixin, ListDropScrollerMixin):
    """Simple list control that provides a drop target and uses
    the new mixin for automatic scrolling.
    """

    def __init__(self, parent, items, get_item_text, columns=None, resize_column=0, allow_drop=True, size=(400,400)):
        if columns is None:
            header_style = wx.LC_NO_HEADER
        else:
            header_style = 0
        wx.ListCtrl.__init__(self, parent, size=size, style=wx.LC_REPORT|header_style)
        listctrl.ListCtrlAutoWidthMixin.__init__(self)
        self.debug = False
        self.get_item_text = get_item_text

        # The mixin needs to be initialized
        ListDropScrollerMixin.__init__(self, interval=200)
        self.set_columns(columns, resize_column)

        if allow_drop:
            self.drop_target = PickledDropTarget(self)
            self.SetDropTarget(self.drop_target)
        self.Bind(wx.EVT_LIST_BEGIN_DRAG, self.on_start_drag)
        self.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)

        self.set_items(items)

    def set_columns(self, columns, resize_column):
        if columns is None:
            columns = ["items"]
        for i, title in enumerate(columns):
            self.InsertColumn(i, title)
        self.setResizeColumn(resize_column)
        self.resizeLastColumn(0)

    def set_items(self, items):
        self.clear(None)
        for item in items:
            self.insert_item(sys.maxsize, item)
        self.resizeLastColumn(0)

    def insert_item(self, index, item):
        index = self.InsertStringItem(index, "placeholder")
        self.set_item_text(index, item)
        self.items[index:index] = [item]

    def set_item_text(self, index, item):
        text = self.get_item_text(item)
        c = self.GetColumnCount()
        if c > 1:
            for i in range(c):
                self.SetStringItem(index, i, text[i])
        else:
            self.SetStringItem(index, 0, text)
        #self.SetItemData(index, sid)

    def on_start_drag(self, evt):
        index = evt.GetIndex()
        #print "beginning drag of item %d" % index

        # Create the data object containing all currently selected
        # items
        data = PickledDataObject()
        items = []
        index = self.GetFirstSelected()
        while index != -1:
            items.append((index, self.items[index]))
            index = self.GetNextSelected(index)
        data.SetData(pickle.dumps((id(self), items),-1))

        # And finally, create the drop source and begin the drag
        # and drop opperation
        drop_source = wx.DropSource(self)
        drop_source.SetData(data)
        #print "Begining DragDrop\n"
        result = drop_source.DoDragDrop(wx.Drag_AllowMove)
        #print "DragDrop completed: %d\n" % result

    def on_context_menu(self, evt):
        menu = wx.Menu()
        menu.Append(wx.ID_SELECTALL, "Select All")
        menu.Append(wx.ID_CLEAR, "Deselect All")
        id = self.GetPopupMenuSelectionFromUser(menu)
        menu.Destroy()
        if id == wx.ID_SELECTALL:
            self.select_all()
        elif id == wx.ID_CLEAR:
            self.deselect_all()

    def add_dropped_items(self, x, y, data):
        src, items = data
        start_index = self.getDropIndex(x, y)
        if self.debug: print(("At (%d,%d), index=%d, adding %s" % (x, y, start_index, items)))
        if start_index == -1:
            start_index = 0

        list_count = self.GetItemCount()

        if src == id(self):
            # reordering items in the same list!
            new_order = list(range(list_count))
            new_indexes = []
            for index, item in items:
                if index < start_index:
                    start_index -= 1
                new_order.remove(index)
                new_indexes.append(index)
            if self.debug: print(("inserting %s into %s at %d" % (str(items), str(new_order), start_index)))
            new_order[start_index:start_index] = new_indexes
            if self.debug: print(("orig list = %s" % str(list(range(list_count)))))
            if self.debug: print((" new list = %s" % str(new_order)))

            self.change_list(new_order, new_indexes)
        else:
            # dropping items from another list
            for _, item in items:
                self.insert_item(start_index, item)
                start_index += 1

    def delete_selected(self):
        list_count = self.GetItemCount()
        new_order = list(range(list_count))
        index = self.GetFirstSelected()
        while index != -1:
            new_order.remove(index)
            index = self.GetNextSelected(index)
        if self.debug: print(("orig list = %s" % str(list(range(list_count)))))
        if self.debug: print((" new list = %s" % str(new_order)))
        self.change_list(new_order)

    @property
    def can_move_up(self):
        return self.GetFirstSelected() > 0

    @property
    def can_move_down(self):
        # There's no convenient way to get all selected items, so as long as
        # the selected item isn't the last one, it can be moved down
        num = self.GetItemCount()
        return num > 0 and self.GetSelectedItemCount() > 0 and not self.IsSelected(num - 1)

    def move_selected(self, delta=-1):
        list_count = self.GetItemCount()
        index_map = [None for i in range(list_count)]
        unselected = set(range(list_count))
        try:
            index = self.GetFirstSelected()
            while index != -1:
                index_map[index + delta] = index
                unselected.remove(index)
                index = self.GetNextSelected(index)
        except IndexError:
            return
        unselected = sorted(unselected)
        new_order = []
        make_selected = []
        for index in range(list_count):
            if index_map[index] is not None:
                new_order.append(index_map[index])
                make_selected.append(index_map[index])
            else:
                new_order.append(unselected.pop(0))
        if self.debug: print(("orig list = %s" % str(list(range(list_count)))))
        if self.debug: print((" new list = %s" % str(new_order)))
        self.change_list(new_order, make_selected)

    def change_list(self, new_order, make_selected=[]):
        """Reorder the list given the new list of indexes.
        
        The new list will be constructed by building up items based on the
        indexes specified in new_order as taken from the current list.  The
        ListCtrl contents is then replaced with the new list of items.  Items
        may also be deleted from the list by not including items in the
        new_order.
        
        @param new_order: list of indexes used to compose the new list
        
        @param make_selected: optional list showing indexes of original order
        that should be marked as selected in the new list.
        """
        new_selection = []
        new_count = len(new_order)
        new_items = []
        for i in range(new_count):
            new_i = new_order[i]
            new_item = self.items[new_i]
            new_items.append(new_item)
            if i != new_i:
                self.set_item_text(i, new_item)

                # save the new highlight position
                if new_i in make_selected:
                    new_selection.append(i)

            # Selection stays with the index even when the item text changes,
            # so remove the selection from all items for the moment
            self.SetItemState(i, 0, wx.LIST_STATE_SELECTED)

        # if the list size has been reduced, clean up any extra items
        list_count = self.GetItemCount()
        for i in range(new_count, list_count):
            self.DeleteItem(new_count)

        # Turn the selection back on for the new positions of the moved items
        for i in new_selection:
            self.SetItemState(i, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)

        self.items = new_items
        if self.debug:
            for i, item in enumerate(self.items):
                print((i, item, self.get_item_text(item)))

    def clear(self, evt=None):
        self.DeleteAllItems()
        self.items = []

    def refresh(self):
        selected = list()  # can't use dict because items might be unhashable
        index = self.GetFirstSelected()
        while index != -1:
            item = self.items[index]
            selected.append(item)
            index = self.GetNextSelected(index)
        items = self.items
        self.set_items(items)
        for i, item in enumerate(self.items):
            if item in selected:  # pay for the slowdown here with O(N^2) search
                self.SetItemState(i, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)

    def select_all(self):
        for i in range(self.GetItemCount()):
            self.SetItemState(i, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)

    def deselect_all(self):
        for i in range(self.GetItemCount()):
            self.SetItemState(i, 0, wx.LIST_STATE_SELECTED)


if __name__ == '__main__':
    class TestList(wx.ListCtrl, ListDropScrollerMixin):
        """Simple list control that provides a drop target and uses
        the new mixin for automatic scrolling.
        """

        def __init__(self, parent, name, count=100):
            wx.ListCtrl.__init__(self, parent, style=wx.LC_REPORT)

            # The mixin needs to be initialized
            ListDropScrollerMixin.__init__(self, interval=200)

            self.drop_target=PickledDropTarget(self)
            self.SetDropTarget(self.drop_target)

            self.create(name, count)

            self.Bind(wx.EVT_LIST_BEGIN_DRAG, self.on_start_drag)

        def create(self, name, count):
            """Set up some test data."""

            self.InsertColumn(0, "#")
            self.InsertColumn(1, "Title")
            for i in range(count):
                self.InsertStringItem(sys.maxsize, str(i))
                self.SetStringItem(i, 1, "%s-%d" % (name, i))

        def on_start_drag(self, evt):
            index = evt.GetIndex()
            print(("beginning drag of item %d" % index))

            # Create the data object containing all currently selected
            # items
            data = PickledDataObject()
            items = []
            index = self.GetFirstSelected()
            while index != -1:
                items.append((self.GetItem(index, 0).GetText(),
                              self.GetItem(index, 1).GetText()))
                index = self.GetNextSelected(index)
            data.SetData(pickle.dumps(items,-1))

            # And finally, create the drop source and begin the drag
            # and drop opperation
            drop_source = wx.DropSource(self)
            drop_source.SetData(data)
            print("Begining DragDrop\n")
            result = drop_source.DoDragDrop(wx.Drag_AllowMove)
            print(("DragDrop completed: %d\n" % result))

        def add_dropped_items(self, x, y, items):
            index = self.getDropIndex(x, y)
            print(("At (%d,%d), index=%d, adding %s" % (x, y, index, items)))

            list_count = self.GetItemCount()
            for item in items:
                if index == -1:
                    index = 0
                index = self.InsertStringItem(index, item[0])
                self.SetStringItem(index, 1, item[1])
                index += 1

        def clear(self, evt):
            self.DeleteAllItems()

    class ListPanel(wx.SplitterWindow):
        def __init__(self, parent):
            wx.SplitterWindow.__init__(self, parent)

            self.list1 = TestList(self, "left", 100)
            self.list2 = TestList(self, "right", 10)
            self.SplitVertically(self.list1, self.list2, 200)
            self.Layout()

    app   = wx.PySimpleApp()
    frame = wx.Frame(None, -1, title='List Drag Test', size=(400,500))
    frame.CreateStatusBar()

    panel = ListPanel(frame)
    label = wx.StaticText(frame, -1, "Drag items from a list to either list.\nThe lists will scroll when the cursor\nis near the first and last visible items")

    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)
    sizer.Add(panel, 1, wx.EXPAND | wx.ALL, 5)
    hsizer = wx.BoxSizer(wx.HORIZONTAL)
    btn1 = wx.Button(frame, -1, "Clear List 1")
    btn1.Bind(wx.EVT_BUTTON, panel.list1.clear)
    btn2 = wx.Button(frame, -1, "Clear List 2")
    btn2.Bind(wx.EVT_BUTTON, panel.list2.clear)
    hsizer.Add(btn1, 1, wx.EXPAND)
    hsizer.Add(btn2, 1, wx.EXPAND)
    sizer.Add(hsizer, 0, wx.EXPAND)

    frame.SetAutoLayout(1)
    frame.SetSizer(sizer)
    frame.Show(1)

    app.MainLoop()



import wx

from .filesystem import find_image_path

default_icon_size = 24

import logging
log = logging.getLogger(__name__)


found_bitmaps = {}

def get_bitmap(action_key, icon_size=0):
    try:
        bitmap = find_bitmap(action_key, icon_size)
    except KeyError:
        try:
            bitmap = find_bitmap("question-mark", icon_size)
        except KeyError:
            bitmap = wx.ArtProvider.GetBitmap(wx.ART_QUESTION, wx.ART_TOOLBAR, (icon_size, icon_size))
        bitmap_key = (action_key, icon_size)
        found_bitmaps[bitmap_key] = bitmap
    return bitmap

def find_bitmap(action_key, icon_size=0):
    bitmap_key = (action_key, icon_size)
    if icon_size < 16:
        icon_size = default_icon_size
    try:
        bitmap = found_bitmaps[bitmap_key]
    except KeyError:
        path = None
        action_key_dashes = action_key.replace("_", "-")
        possibilities = [action_key, f"{action_key}-{icon_size}", f"icons8-{action_key}-{icon_size}", f"icons8-{action_key_dashes}-{icon_size}"]
        for prefix in possibilities:
            try:
                path = find_image_path(prefix + ".png")
            except OSError:
                continue

        if path is not None:
            log.debug(f"find_bitmap: using {path} for {action_key}")
            bitmap = wx.Bitmap(wx.Image(path))
            found_bitmaps[bitmap_key] = bitmap
        else:
            log.warning(f"find_bitmap: no icon for {action_key}; checked prefixes {possibilities}")
            raise KeyError
    return bitmap



# Standard library imports.
import os
import time
import wx
import wx.lib.newevent

from ..errors import ProgressCancelError

import logging
log = logging.getLogger(__name__)

# wx logging handler from http://stackoverflow.com/questions/2819791/how-can-i-redirect-the-logger-to-a-wxpython-textctrl-using-a-custom-logging-hand
# create event type
wxLogEvent, EVT_WX_LOG_EVENT = wx.lib.newevent.NewEvent()


class ProgressDialog(wx.Dialog):
    def __init__(self, parent, title="Progress", delay=1000):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=wx.DefaultSize, pos=wx.DefaultPosition,
                           style=wx.DEFAULT_DIALOG_STYLE)
        self.border = 20
        self.delay = delay

        sizer = wx.BoxSizer(wx.VERTICAL)

        self.label = wx.StaticText(self, -1, "Working...")
        self.label.SetMinSize((400, -1))
        sizer.Add(self.label, 0, flag=wx.EXPAND|wx.ALL, border=self.border)

        self.progressHeight = 12
        self.gauge = wx.Gauge(self, -1,
              range=100, size = (-1, self.progressHeight),
              style=wx.GA_HORIZONTAL|wx.GA_SMOOTH)
        sizer.Add(self.gauge, 0, flag=wx.EXPAND|wx.ALL, border=self.border)

        #self.finished = wx.lib.expando.ExpandoTextCtrl(self, -1, value="")
        self.finished = wx.TextCtrl(self, -1, value="", size=(-1,100), style=wx.TE_MULTILINE)
        self.finished.SetEditable(False)
        #self.finished.SetMaxHeight(200)
        self.finished.Bind(wx.lib.expando.EVT_ETC_LAYOUT_NEEDED, self.on_refit)
        sizer.Add(self.finished, 0, flag=wx.EXPAND|wx.ALL, border=self.border)

        self.count = 0
        self.last_pulse = time.clock()
        self.time_delta = 0.01

        self.visible = False
        self._delaytimer = wx.PyTimer(self.on_timer)
        self.is_pulse = True

        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_CANCEL)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 0, flag=wx.EXPAND|wx.ALL, border=5)
        btn.Bind(wx.EVT_BUTTON, self.on_cancel)
        self.request_cancel = False

        self.finished.Hide()
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.label.Layout()
        self.gauge.Layout()
        self.Layout()
        self.CenterOnParent()

    def on_cancel(self, evt):
        self.request_cancel = True

    def on_refit(self, evt):
        # Note that there is a bug in expando that resets height to one line
        # after the SetMaxHeight is reached.
        self.Fit()

    def raise_error(self):
        raise ProgressCancelError(self.GetTitle() + " canceled by user!")

    def start_visibility_timer(self):
        if not self._delaytimer.IsRunning():
            self._delaytimer.Start(self.delay, oneShot=True)

    def stop_visibility_timer(self):
        if self._delaytimer.IsRunning():
            self._delaytimer.Stop()

    def on_timer(self):
        self.Show()

    def add_finished(self, text, time):
        self.finished.Show()
        self.Fit()
        self.Layout()
        t = self.finished.GetValue()
        if t:
            text = "\n" + text
        self.finished.AppendText("%s: %fs" % (text, time))

    def set_ticks(self, count):
        """Set the total number of ticks that will be contained in the
        progress bar.
        """
        self.gauge.SetRange(count)
        self.count = 0
        self.last_pulse = 0 # force update next time
        self.is_pulse = False

    def is_update_time(self):
        t = time.clock()
        if t - self.last_pulse > self.time_delta:
            self.last_pulse = t
            return True
        return False

    def set_pulse(self):
        """Set the total number of ticks that will be contained in the
        progress bar.
        """
        if self.is_update_time():
            self.gauge.Pulse()
        self.is_pulse = True
        wx.Yield()

    def tick(self, text=None, index=None):
        """Advance the progress bar by one tick and update the label.
        """
        ok = self.is_update_time()
        if self.is_pulse:
            if text:
                self.label.SetLabel(text)
            if ok:
                self.gauge.Pulse()
        else:
            if text:
                self.label.SetLabel(text)
            if index is None:
                self.count += 1
            else:
                self.count = index
            if self.count > self.gauge.GetRange():
                self.set_pulse()
            elif ok:
                self.gauge.SetValue(self.count)
        self.gauge.Update()
        wx.Yield()


class wxLogHandler(logging.Handler):
    """
    A handler class which sends log strings to a wx object
    """

    progress_dialog = None

    disabler = None

    def __init__(self, default_title=""):
        """
        Initialize the handler
        @param wxDest: the destination object to post the event to 
        @type wxDest: wx.Window
        """
        logging.Handler.__init__(self)
        self.level = logging.DEBUG
        self.default_title = default_title
        self.use_gui = True
        self.time_t0 = 0

    def flush(self):
        """
        does nothing for this handler
        """

    def emit(self, record):
        """
        Emit a record.

        """
        msg = self.format(record)

        # Handle progress cancel request here, the only place that's not inside
        # a wx event handler.  Attempting to handle inside an event handler
        # doesn't propagate outside the event handler.
        d = self.__class__.progress_dialog
        if d and d.request_cancel and msg != "END":
            # change flag so multiple requests are not processed
            d.request_cancel = False
            raise ProgressCancelError(d.GetTitle() + " canceled by user!")

        try:
            evt = wxLogEvent(message=msg,levelname=record.levelname)
            self.post(evt)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    @classmethod
    def get_dialog(cls):
        if cls.progress_dialog is None or not cls.progress_dialog:
            top = wx.GetApp().GetTopWindow()
            cls.progress_dialog = ProgressDialog(top)
        return cls.progress_dialog

    @classmethod
    def get_dialog_if_open(cls):
        if cls.progress_dialog is None or not cls.progress_dialog:
            return None
        return cls.progress_dialog

    @classmethod
    def open_dialog(cls):
        d = cls.get_dialog()
        # Forcibly disable all windows (other than the progress dialog) to
        # prevent user event processing in the wx.Yield calls.
        cls.disabler = wx.WindowDisabler(d)
        wx.BeginBusyCursor()
        return d

    @classmethod
    def close_dialog(cls):
        d = cls.get_dialog_if_open()
        if d is not None:
            cls.disabler = None
            d.stop_visibility_timer()
            d.Destroy()
            wx.Yield()
            wx.EndBusyCursor()  # fails in wx 3.0 before Yield
        cls.progress_dialog = None

    def force_cursor(self):
        # OS X resets the busy cursor when the cursor moves out of the dialog,
        # so at every tick call this method to reset it to the wait cursor.
        # Other platforms don't have this problem.
        wx.SetCursor(wx.Cursor(wx.CURSOR_WAIT))

    def post(self, evt):
        if not self.use_gui:
            print("NO GUI: message=%s" % evt.message)
            return
        m = evt.message
        if m.startswith("START"):
            self.time_t0 = time.clock()
            d = self.open_dialog()
            if "=" in m:
                _, text = m.split("=", 1)
                d.SetTitle(text)
            else:
                d.SetTitle(self.default_title)
            d.start_visibility_timer()
            wx.Yield()
        elif m == "END":
            self.close_dialog()
        elif m == "NO GUI":
            self.close_dialog()
            self.use_gui = False
        else:
            d = self.get_dialog_if_open()
            if d is None:
                # skipping log message to dialog if no dialog is open
                return
            if m.startswith("TITLE"):
                self.force_cursor()
                _, text = m.split("=", 1)
                d.SetTitle(text)
                wx.Yield()
            elif m.startswith("TICKS"):
                self.force_cursor()
                _, count = m.split("=", 1)
                d.set_ticks(int(count))
            elif m.startswith("TICK"):
                self.force_cursor()
                if "=" in m:
                    _, count = m.split("=", 1)
                    d.tick(index=int(count))
                else:
                    d.tick()
            elif m == "PULSE":
                self.force_cursor()
                d.set_pulse()
            elif m.startswith("TIME_DELTA"):
                t = time.clock()
                self.force_cursor()
                _, text = m.split("=", 1)
                d.add_finished(text, t - self.time_t0)
                wx.Yield()
                self.time_t0 = t
            else:
                self.force_cursor()
                d.tick(m)


def is_active():
    return wxLogHandler.get_dialog_if_open() is not None


def attach_handler():
    log = logging.getLogger("progress")
    level = log.getEffectiveLevel()
    if level > logging.INFO:
        log.setLevel(logging.INFO)
    log.propagate = False
    handler = wxLogHandler("Progress")
    log.addHandler(handler)


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    def progress_test():
        print("HI!")
        progress_log.info("START=First Test")
        try:
            progress_log.info("TITLE=Starting timer")
            for i in range(1000):
                print(i)
                progress_log.info("Trying %d" % i)
                for j in range(10):
                    time.sleep(.01)
                    wx.Yield()
                if i > 4:
                    progress_log.info("TIME_DELTA=Finished trying %d" % i)
                progress_log.info("PULSE")
                wx.Yield()

        except ProgressCancelError as e:
            error = str(e)
        finally:
            progress_log.info("END")

    p = FileProgressPlugin()
    p.start()
    progress_log = logging.getLogger("progress")

    app = wx.App(redirect = False)
    frame = wx.Frame(None, title='Progress Test', size=(800,400))
    frame.Show()
    wx.CallLater(1000, progress_test)
    app.MainLoop()



import sys
import functools
import weakref

import wx

from sawx.utils.command import DisplayFlags

try:
    from atrcopy import match_bit_mask, comment_bit_mask, user_bit_mask, selected_bit_mask, diff_bit_mask
except ImportError:
    user_bit_mask = 0x07
    diff_bit_mask = 0x10
    match_bit_mask = 0x20
    comment_bit_mask = 0x40
    selected_bit_mask = 0x80

import logging
logging.basicConfig()
log = logging.getLogger(__name__)
# log.setLevel(logging.DEBUG)
caret_log = logging.getLogger("caret")
# caret_log.setLevel(logging.DEBUG)
mode_log = logging.getLogger("mouse_mode")
# mode_log.setLevel(logging.DEBUG)

##### Carets

@functools.total_ordering
class Caret:
    """Class representing both a caret's row/col position, and optionally a
    single selected range or rectangular region.

    Anchor indexes behave like caret positions: they indicate positions
    between bytes.
    """
    def __init__(self, row=0, col=0, rectangular=False, state=None):
        if state is not None:
            self.restore(state)
        else:
            try:
                row.anchor_start
                self.restore(row.serialize())
            except AttributeError:
                self.rc = (row, col)
                self.anchor_start = self.anchor_initial_start = self.anchor_end = self.anchor_initial_end = (-1, -1)
                self.rectangular = rectangular

    def __bool__(self):
        return self.rc is not None

    __nonzero__=__bool__

    def __eq__(self, other):
        if not hasattr(self, 'rc'):
            caret_log.error("not a Caret object")
            return False
        caret_log.debug("comparing caret positions: %d %d" % (self.rc, other.rc))
        return self.rc == other.rc

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if self.has_selection:
            if other.has_selection:
                return self.anchor_start < other.anchor_start
            return self.anchor_start < other.rc
        return self.rc < other.rc

    def __repr__(self):
        return "Caret%s" % str(self.serialize())

    def __hash__(self):
        return hash(self.serialize())

    @property
    def has_selection(self):
        return self.anchor_start >= (0,0)

    @property
    def range(self):
        if not self.has_selection:
            raise ValueError("No selection")
        s = self.anchor_start
        e = self.anchor_end
        return (s, e) if e > s else (e, s)

    @property
    def range_including_caret(self):
        if self.has_selection:
            s = self.anchor_start
            e = self.anchor_end
            if s > e: s, e = e, s
        else:
            s = e = self.rc
        return (s, e)

    def set(self, r, c):
        self.rc = (r, c)
        self.clear_selection()

    def clear_selection(self):
        self.anchor_start = self.anchor_initial_start = self.anchor_end = self.anchor_initial_end = (-1, -1)

    def set_initial_selection(self, start, end):
        self.anchor_start = self.anchor_initial_start = start
        self.anchor_end = self.anchor_initial_end = end

    def set_selection(self, start, end):
        self.anchor_start = start
        self.anchor_end = end

    def serialize(self):
        return (self.rc, self.anchor_start, self.anchor_initial_start, self.anchor_end, self.anchor_initial_end, self.rectangular)

    def restore(self, state):
        try:
            self.rc, self.anchor_start, self.anchor_initial_start, self.anchor_end, self.anchor_initial_end, self.rectangular = state
        except TypeError:
            self.rc = state
            self.clear_selection()

    def copy(self):
        state = self.serialize()
        return Caret(state=state)

    def contains(self, other):
        return self.anchor_start >= (0,0) and other.anchor_start >= self.anchor_start and other.anchor_end <= self.anchor_end

    def intersects(self, other):
        return not (other.anchor_start > self.anchor_end or other.anchor_end < self.anchor_start)

    def in_selection(self, row, col):
        return (row, col) >= self.anchor_start and (row, col) <= self.anchor_end

    def merge(self, other):
        """Merge boundaries of other into this caret"""
        if other.anchor_start < self.anchor_start:
            self.anchor_start = other.anchor_start
            self.anchor_initial_start = other.anchor_initial_start
            if other.rc < self.rc:
                self.rc = other.rc
        if other.anchor_end > self.anchor_end:
            self.anchor_end = other.anchor_end
            self.anchor_initial_end = other.anchor_initial_end
            if other.rc > self.rc:
                self.rc = other.rc

    def add_selection_to_style(self, table):
        if self.has_selection:
            if self.rectangular:
                pass
            else:
                start, end = self.anchor_start, self.anchor_end
                if end < start:
                    start, end = end, start
                index1, _ = table.get_index_range(*start)
                _, index2 = table.get_index_range(*end)
                table.set_selected_index_range(index1, index2)


class MultiCaretHandler:
    def __init__(self):
        self.carets = []

    def __str__(self):
        return "MultiCaretHandler: carets=" + " ".join([str(c) for c in self.carets])

    def __len__(self):
        return len(self.carets)

    @property
    def current(self):
        return self.carets[-1]  # last one is the most recent

    @property
    def has_carets(self):
        return len(self.carets) > 0

    @property
    def has_selection(self):
        for caret in self.carets:
            if caret.has_selection:
                return True
        return False

    @property
    def carets_with_selection(self):
        for caret in self.carets:
            if caret.has_selection:
                yield caret

    @property
    def selected_ranges(self):
        return [c.range for c in self.carets if c.has_selection]

    def copy(self):
        handler = MultiCaretHandler()
        handler.carets = [c.copy() for c in self.carets]
        return handler

    def new_carets(self, caret_state):
        self.carets = [Caret(state=s) for s in caret_state]

    def calc_state(self):
        return [caret.serialize() for caret in self.carets]

    def has_changed_state(self, other_state):
        current = self.calc_state()
        return current == other_state

    def convert_to_indexes(self, table):
        index_list = []
        for caret in self.carets:
            index, _ = table.get_index_range(*caret.rc)
            if caret.anchor_start[0] < 0:
                anchor_start = -1
            else:
                anchor_start, _ = table.get_index_range(*caret.anchor_start)
            if caret.anchor_end[0] < 0:
                anchor_end = -1
            else:
                anchor_end, _ = table.get_index_range(*caret.anchor_end)
            index_list.append((index, anchor_start, anchor_end))
        return index_list

    def convert_from_indexes(self, table, indexes):
        self.carets = []
        for index, anchor_start, anchor_end in indexes:
            r, c = table.index_to_row_col(index)
            caret = Caret(r, c)
            r, c = table.index_to_row_col(anchor_start)
            caret.anchor_start = self.anchor_initial_start = (r, c)
            r, c = table.index_to_row_col(anchor_end)
            caret.anchor_end = self.anchor_initial_end = (r, c)
            self.carets.append(caret)

    def add_caret(self, caret):
        self.carets.append(caret)

    def move_carets_vertically(self, table, delta_r):
        caret_log.debug(f"moving vertically: {delta_r}")
        for caret in self.carets:
            caret.rc = table.enforce_valid_row_col(caret.rc[0] + delta_r, caret.rc[1])
            caret.clear_selection()
        self.validate_carets()

    def move_carets_horizontally(self, table, delta_c, wrap=False):
        caret_log.debug(f"moving horizontally: {delta_c}")
        for caret in self.carets:
            r = caret.rc[0]
            c = caret.rc[1] + delta_c
            if wrap and not table.is_row_col_inside(r, c):
                if c < 0:
                    r -= 1
                    c += table.get_items_in_row(r)
                else:
                    r += 1
                    c -= table.get_items_in_row(r)
            caret.rc = table.enforce_valid_row_col(r, c)
            caret.clear_selection()
        self.validate_carets()

    def move_carets_to(self, r, c):
        self.carets = [Caret(r, c)]

    def move_current_caret_to(self, r, c):
        try:
            caret = self.carets[-1]
            caret.rc = r, c
        except IndexError:
            self.move_carets_to(r, c)

    def move_current_caret_to_index(self, table, index):
        r, c = table.index_to_row_col(index)
        self.carets = [Caret(r, c)]

    def move_carets_process_function(self, func):
        for caret in self.carets:
            caret.rc = func(*caret.rc)
            caret.clear_selection()

    def validate_carets(self):
        # new_carets = []
        # for caret in self.carets:
        #     index = self.validate_caret_position(index)
        #     new_carets.append(index)
        # self.carets = new_carets
        pass

    def validate_caret_position(self, table, index):
        return table.enforce_valid_index(index)

    def refresh_style_from_selection(self, table):
        table.clear_selected_style()
        for caret in self.carets:
            caret.add_selection_to_style(table)

    def collapse_overlapping(self):
        """Check if the current caret selection overlaps any existing caret and
        merge any overlaps into the current caret.
        """
        caret_log.debug("before collapsed carets: {str(self.carets)}")
        try:
            current = self.carets.pop()
        except IndexError:
            # No carets! Can't be an overlap
            return
        if self.carets:
            if current.has_selection:
                collapsed = []
                for c in self.carets:
                    if current.contains(c):
                        # gets rid of any selections wholly contained within
                        # the current selection, or any standalone carets
                        # inside the selection
                        continue
                    elif current.intersects(c):
                        current.merge(c)
                    else:
                        collapsed.append(c)
                self.carets = collapsed
            else:
                # Merge caret into selection
                collapsed = []
                for c in self.carets:
                    if current.has_selection:
                        # if caret has been merged into a selection, don't do
                        # any further checking.
                        collapsed.append(c)
                    else:
                        if not c.has_selection and c.rc == current.rc:
                            # gets rid of any duplicate carets
                            continue
                        elif c.contains(current):
                            # if caret inside a selection, make the that the
                            # current caret
                            current = c
                        else:
                            collapsed.append(c)
                self.carets = collapsed
        self.carets.append(current)
        caret_log.debug(f"collapsed carets {str(self.carets)}")

    def collapse_selections_to_carets(self):
        for caret in self.carets:
            caret.clear_selection()

    def get_selected_ranges_and_indexes(self, table):
        opt = self.get_selected_ranges(table)
        return opt, ranges_to_indexes(opt)

    def invert_selection_ranges(self, table, ranges):
        return invert_ranges(ranges, caret_handler.document_length)

    def process_char_flags(self, flags):
        """Perform the UI updates described by the DisplayFlags instance
        
        """
        visible_range = False
        caret_moved = False
        log.debug("processing caret flags: %s" % str(flags))

        if flags.old_carets is not None:
            log.debug(f"old_carets: {flags.old_carets}")
            if flags.add_caret:
                self.carets.add_old_carets(flags.old_carets)
                caret_moved = True
                log.debug("caret added! before: %s, after: %s" % (flags.old_carets, self.carets))
            elif flags.force_single_caret:
                if flags.caret_index is not None:
                    self.carets.new_caret(flags.caret_index)
                self.carets.remove_old_carets()
                caret_moved = True
                log.debug("force_single_caret")
            else:
                self.validate_carets()
                caret_state = self.calc_state()
                caret_moved = caret_state != flags.old_carets
                log.debug("caret moved: %s old_carets: %s, new carets: %s" % (caret_moved, flags.old_carets, caret_state))
            if caret_moved:
                if not flags.keep_selection:
                    index = self.current.index
                    self.current.set_initial_selection(index, index)
                visible_range = True
                # self.sync_caret_event = flags
                log.debug("caret_moved")
        elif flags.force_single_caret:
            log.debug(f"force_single_caret: caret_index={flags.caret_index}")
            if flags.caret_index is not None:
                c = Caret(flags.caret_index)
                self.carets.force_single_caret(c)
                caret_moved = True
                # self.sync_caret_event = flags

        if flags.index_range is not None:
            log.debug(f"index_range: {flags.index_range}")
            if flags.select_range:
                log.debug(f"select_range")
                self.current.set_anchors(flags.index_range[0], flags.index_range[1])
                document.change_count += 1
            visible_range = True

        if visible_range:
            # Only update the range on the current editor, not other views
            # which are allowed to remain where they are
            log.debug(f"visible_range: index_visible={flags.index_visible}")
            if flags.index_visible is None:
                flags.index_visible = self.current.index
            self.ensure_visible_event = flags

            flags.refresh_needed = True

        if flags.viewport_origin is not None:
            flags.source_control.move_viewport_origin(flags.viewport_origin)
            flags.skip_source_control_refresh = True
            flags.refresh_needed = True
        log.debug(f"FINISHED process_caret_flags: refresh_needed={flags.refresh_needed}, skip_source_control_refresh={flags.skip_source_control_refresh}")


##### Mouse modes

class MouseMode:
    """
    Processing of mouse events, separate from the rendering window
    
    This is an object-based control system of mouse modes
    """
    icon = "help.png"
    menu_item_name = "Generic Mouse Mode"
    menu_item_tooltip = "Tooltip for generic mouse mode"

    mouse_too_close_pixel_tolerance = 5

    def __init__(self, window):
        self.control = weakref.proxy(window)
        self.snapped_point = None, 0
        self.first_mouse_down_position = 0, 0
        self.after_first_mouse_up = False
        self.mouse_up_too_close = False
        self.can_snap = False

        # left_up events can be called after double click on some platforms, so
        # this parameter can be set by the parent of the handler if the handler
        # wants to differentiate response to left_up
        self.num_clicks = 0

        # Optional (only OS X at this point) mouse wheel event filter
        self.wheel_scroll_count = 0
        self.use_every_nth_wheel_scroll = 5

        self.init_post_hook()

    def init_post_hook(self):
        pass

    def cleanup(self):
        pass

    def get_cursor(self):
        return wx.Cursor(wx.CURSOR_ARROW)

    def get_position(self, evt):
        return evt.GetPosition()

    def reset_early_mouse_params(self):
        self.mouse_up_too_close = False
        self.after_first_mouse_up = False

    def check_early_mouse_release(self, evt):
        c = self.control
        p = evt.GetPosition()
        dx = p[0] - self.first_mouse_down_position[0]
        dy = p[1] - self.first_mouse_down_position[1]
        tol = self.mouse_too_close_pixel_tolerance
        if abs(dx) < tol and abs(dy) < tol:
            return True
        return False

    ##### Sublasses need to override these for customized actions

    def process_mouse_motion_down(self, evt):
        evt.Skip()

    def process_left_down(self, evt):
        evt.Skip()

    def process_mouse_motion_up(self, evt):
        evt.Skip()

    def process_left_up(self, evt):
        evt.Skip()

    def process_left_dclick(self, evt):
        evt.Skip()

    def process_popup(self, evt):
        data = self.calc_popup_data(evt)
        self.control.add_popup_data(evt, data)
        log.debug(f"process_popup: found popup data {data}")
        popup_desc = self.calc_popup_menu(evt)
        if popup_desc:
            log.debug(f"process_popup: found popup menu for mode {self.menu_item_name}; skipping calc_popup_menu from {self.control}")
        else:
            popup_desc = self.control.calc_popup_menu(evt)
        log.debug(f"process_popup: using popup menu {popup_desc}")
        if popup_desc:
            self.show_popup(popup_desc, data)

    def calc_popup_data(self, evt):
        return {}

    def calc_popup_menu(self, evt):
        return []

    def show_popup(self, popup_desc, data):
        # default is a proxy to the CompactGrid object, but subclasses may have
        # different needs depending on the mode, so this hook is provided.
        self.control.show_popup(popup_desc, data)

    def process_mouse_wheel(self, evt):
        c = self.control
        rotation = evt.GetWheelRotation()
        delta = evt.GetWheelDelta()
        window = evt.GetEventObject()
        mode_log.debug("on_mouse_wheel_scroll. rot=%s delta=%d win=%s" % (rotation, delta, window))
        if rotation == 0 or delta == 0:
            return

        if sys.platform == "darwin":
            # OS X mouse wheel handling is not the same as other platform.
            # The delta value is 10 while the other platforms are 120,
            # and the rotation amount varies, seemingly due to the speed at
            # which the wheel is rotated (or how fast the trackpad is swiped)
            # while other platforms are either 120 or -120.  When mouse wheel
            # handling is performed in the usual manner on OS X it produces a
            # strange back-and-forth zooming in/zooming out.  So, this extra
            # hack is needed to operate like the other platforms.

            # add extra to the rotation so the minimum amount is 1 or -1
            extra = delta if rotation > 0 else -delta
            amount = (rotation + extra) // delta
            self.wheel_scroll_count -= abs(amount)
            if self.wheel_scroll_count > 0:
                return
            self.wheel_scroll_count = self.use_every_nth_wheel_scroll
        else:
            amount = rotation // delta

        if evt.ControlDown():
            if amount < 0:
                self.zoom_out(evt, amount)
            elif amount > 0:
                self.zoom_in(evt, amount)
        elif not evt.ShiftDown() and not evt.AltDown():
            self.pan_mouse_wheel(evt, amount)
        else:
            evt.Skip()

    def zoom_out(self, evt, amount):
        evt.Skip()

    def zoom_in(self, evt, amount):
        evt.Skip()

    def pan_mouse_wheel(self, evt, amount):
        self.control.pan_mouse_wheel(evt)

    def process_mouse_enter(self, evt):
        evt.Skip()

    def process_mouse_leave(self, evt):
        evt.Skip()

    def process_focus(self, evt):
        pass

    def process_focus_lost(self, evt):
        pass

    #### Selection

    def select_all(self, caret_handler, refresh=True):
        """ Selects the entire document
        """
        caret_handler.move_carets_to(0, 0)
        caret = caret_handler.current
        start = caret.rc
        t = self.control.table
        end = t.index_to_row_col(t.last_valid_index)
        caret.set_initial_selection(start, end)
        self.highlight_selected_ranges(caret_handler)
        self.control.update_ui_for_selection_change()

    def select_none(self, caret_handler, refresh=True):
        """ Clears any selection in the document
        """
        caret_handler.collapse_selections_to_carets()
        self.highlight_selected_ranges(caret_handler)
        self.control.update_ui_for_selection_change()

    def select_none_if_selection(self, caret_handler):
        if caret_handler.has_selection:
            self.select_none(caret_handler)

    def select_ranges(self, caret_handler, ranges, refresh=True):
        """ Selects the specified ranges
        """
        caret_handler.carets.selected_ranges = ranges
        try:
            start, end = caret_handler.carets.selected_ranges[-1]
        except IndexError:
            start, end = 0, 0
        caret_handler.anchor_start_index = caret_handler.anchor_initial_start_index = start
        caret_handler.anchor_end_index = caret_handler.anchor_initial_end_index = end
        self.highlight_selected_ranges(caret_handler)
        self.control.update_ui_for_selection_change()

    def select_invert(self, caret_handler, refresh=True):
        """ Selects the entire document
        """
        ranges = self.invert_selection_ranges(caret_handler, caret_handler.carets.selected_ranges)
        self.select_ranges(caret_handler, ranges, refresh)

    def select_range(self, caret_handler, start, end, add=False, extend=False):
        """ Adjust the current selection to the new start and end indexes
        """
        if extend:
            caret = caret_handler.current
            caret.set_selection(start, end)
        elif add:
            caret = Caret(end)
            caret.set_initial_selection(start, end)
            caret_handler.add_caret(caret)
        else:
            caret = Caret(end)
            caret.set_initial_selection(start, end)
            caret_handler.carets.force_single_caret(caret)
        self.refresh_ranges(caret_handler)

    def refresh_ranges(self, caret_handler):
        log.debug("refreshing ranges: %s" % str(caret_handler.carets))
        self.highlight_selected_ranges(caret_handler)
        self.control.update_ui_for_selection_change()

    def highlight_selected_ranges(self, caret_handler):
        # log.error("highlight_selected_ranges not defined in mouse mode")
        caret_handler.refresh_style_from_selection(self.control.table)


class NormalSelectMode(MouseMode):
    def init_post_hook(self):
        self.last_mouse_event = -1, -1
        self.event_modifiers = None

    def process_mouse_motion_up(self, evt):
        mode_log.debug("NormalSelectMode: process_mouse_motion_up")
        cg = self.control
        input_row, input_cell = cg.get_row_cell_from_event(evt)
        if (input_row, input_cell) == self.last_mouse_event:
            # only process if mouse has moved to a new cell; no sub-cell
            # events!
            return
        self.last_mouse_event = (input_row, input_cell)
        cmd = self.calc_mouse_motion_up_command(evt, input_row, input_cell)
        if cmd:
            cg.process_command(cmd)

    def calc_mouse_motion_up_command(self, evt, row, cell):
        cg = self.control
        col = cg.main.cell_to_col(row, cell)
        cg.handle_motion_update_status(evt, row, col)

    def process_left_down(self, evt):
        cg = self.control
        flags = cg.create_mouse_event_flags()
        input_row, input_cell = cg.get_row_cell_from_event(evt)
        mode_log.debug(f"NormalSelectMode: process_left_down: {input_row} {input_cell}")
        self.event_modifiers = evt.GetModifiers()
        self.last_mouse_event = (input_row, input_cell)
        cmd = self.calc_left_down_command(evt, input_row, input_cell, flags)
        if cmd:
            cg.process_command(cmd)
        else:
            cg.Refresh()

    def calc_left_down_command(self, evt, row, cell, flags):
        cg = self.control
        cg.main.update_caret_from_mouse(row, cell, flags)
        cg.handle_select_start(evt, cg.main.current_caret_row, cg.main.current_caret_col, flags)

    def process_mouse_motion_down(self, evt):
        cg = self.control
        input_row, input_cell = cg.get_row_cell_from_event(evt)
        if (input_row, input_cell) == self.last_mouse_event:
            # only process if mouse has moved to a new cell; no sub-cell
            # events!
            return
        mode_log.debug(f"NormalSelectMode: process_mouse_motion_down: {input_row} {input_cell}")
        self.last_mouse_event = (input_row, input_cell)
        flags = cg.create_mouse_event_flags()
        cmd = self.calc_mouse_motion_down_command(evt, input_row, input_cell, flags)
        if cmd:
            cg.process_command(cmd)
        else:
            cg.Refresh()

    def calc_mouse_motion_down_command(self, evt, row, cell, flags):
        cg = self.control
        last_row, last_col = cg.main.current_caret_row, cg.main.current_caret_col
        cg.main.handle_user_caret(row, cell, flags)
        if last_row != cg.main.current_caret_row or last_col != cg.main.current_caret_col:
            cg.handle_select_motion(evt, cg.main.current_caret_row, cg.main.current_caret_col, flags)

    def process_left_up(self, evt):
        mode_log.debug("NormalSelectMode: process_left_up")
        cg = self.control
        cg.main.scroll_timer.Stop()
        self.event_modifiers = None
        cg.handle_select_end(evt, cg.main.current_caret_row, cg.main.current_caret_col)

    def process_left_dclick(self, evt):
        mode_log.debug("NormalSelectMode: process_left_dclick")
        evt.Skip()

    def calc_popup_data(self, evt):
        cg = self.control
        row, col, inside = cg.get_row_col_from_event(evt)
        if inside:
            index, _ = cg.table.get_index_range(row, col)
            style = cg.table.segment.style[index]
        else:
            style = 0
        popup_data = {
            'control': cg,
            'index': index,
            'in_selection': style&0x80,
            'row': row,
            'col': col,
            'inside': inside,
            }
        return popup_data

    def zoom_in(self, evt, amount):
        self.control.zoom_in()

    def zoom_out(self, evt, amount):
        self.control.zoom_out()


class RectangularSelectMode(NormalSelectMode):
    icon = "select.png"
    menu_item_name = "Select"
    menu_item_tooltip = "Select rectangular regions"

    def display_coords(self, evt, extra=None):
        mode_log.debug("display_coords")
        cg = self.control
        v = cg.segment_viewer
        row, col, inside = cg.get_row_col_from_event(evt)
        index, _ = cg.table.get_index_range(row, col)
        msg = "x=$%x y=$%x index=$%x" % (col, row, index)
        if extra:
            msg += " " + extra
        cg.show_status_message(msg)


class GridCellTextCtrl(wx.TextCtrl):
    def __init__(self, parent, id, num_chars_autoadvance, *args, **kwargs):
        # Don't use the validator here, because apparently we can't
        # reset the validator based on the columns.  We have to do the
        # validation ourselves using EVT_KEY_DOWN.
        wx.TextCtrl.__init__(self, parent, id, *args, style=wx.TE_PROCESS_TAB|wx.TE_PROCESS_ENTER, **kwargs)
        log.debug("parent=%s" % parent)
        self.SetInsertionPoint(0)
        self.Bind(wx.EVT_TEXT, self.on_text)
        self.Bind(wx.EVT_CHAR, self.on_char)
        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)
        self.SetMaxLength(num_chars_autoadvance)
        self.num_chars_autoadvance = num_chars_autoadvance

    def is_valid_keycode(self, keycode):
        return True

    def on_key_down(self, evt):
        """
        Keyboard handler to process command keys before they are
        inserted.  Tabs, arrows, ESC, return, etc. should be handled
        here.  If the key is to be processed normally, evt.Skip must
        be called.  Otherwise, the event is eaten here.

        @param evt: key event to process
        """
        log.debug("key down before evt=%s" % evt.GetKeyCode())
        key = evt.GetKeyCode()

        if key == wx.WXK_TAB:
            wx.CallAfter(self.GetParent().advance_caret)
        elif key == wx.WXK_ESCAPE:
            wx.CallAfter(self.GetParent().end_editing)
        elif key == wx.WXK_RETURN:
            wx.CallAfter(self.GetParent().accept_edit, self.num_chars_autoadvance)
        else:
            evt.Skip()

    def on_char(self, evt):
        code = evt.GetKeyCode()
        char = evt.GetUnicodeKey()
        log.debug("char keycode=%s unicode=%s" % (code, char))
        if char == wx.WXK_NONE or code == wx.WXK_BACK or self.is_valid_keycode(code):
            evt.Skip()
            wx.CallAfter(self.GetParent().Refresh)

    def get_processed_value(self):
        return self.GetValue()

    def on_text(self, evt):
        """
        Callback used to automatically advance to the next edit field. If
        self.num_chars_autoadvance > 0, this number is used as the max number
        of characters in the field.  Once the text string hits this number, the
        field is processed and advanced to the next position.
        """
        log.debug("evt=%s str=%s cursor=%d" % (evt, evt.GetString(), self.GetInsertionPoint()))

        # NOTE: we check that GetInsertionPoint returns 1 less than
        # the desired number because the insertion point hasn't been
        # updated yet and won't be until after this event handler
        # returns.
        n = self.num_chars_autoadvance
        if n and len(evt.GetString()) >= n and self.GetInsertionPoint() >= n - 1:
            # FIXME: problem here with a bunch of really quick
            # keystrokes -- the interaction with the
            # underlyingSTCChanged callback causes a cell's
            # changes to be skipped over.  Need some flag in grid
            # to see if we're editing, or to delay updates until a
            # certain period of calmness, or something.
            log.debug("advancing after edit")
            wx.CallAfter(self.GetParent().accept_edit, n)





class MouseEventMixin:
    default_mouse_mode_cls = NormalSelectMode

    def __init__(self, caret_handler, mouse_mode_cls=None):
        self.caret_handler = caret_handler
        self.multi_select_mode = False
        self.select_extend_mode = False
        self.mouse_drag_started = False
        self.is_editing_in_cell = False
        self.is_mousing_while_editing = False
        self.pending_select_awaiting_drag = None
        self.source = None
        self.edit_source = None
        self.forced_cursor = None
        self.batch = None

        if mouse_mode_cls is None:
            mouse_mode_cls = self.default_mouse_mode_cls
        self.default_mouse_mode_cls = mouse_mode_cls  # override class attr
        self.default_mouse_mode = self.default_mouse_mode_cls(self)
        self.mouse_mode = self.default_mouse_mode  # can't call set_mouse_mode yet because control hasn't been initialized

    def map_mouse_events(self, source):
        self.source = source
        source.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)
        source.Bind(wx.EVT_LEFT_DOWN, self.on_left_down)
        source.Bind(wx.EVT_MOTION, self.on_motion)
        source.Bind(wx.EVT_LEFT_UP, self.on_left_up)
        source.Bind(wx.EVT_LEFT_DCLICK, self.on_left_dclick)
        source.Bind(wx.EVT_RIGHT_DOWN, self.on_popup)
        source.Bind(wx.EVT_MOUSEWHEEL, self.on_mouse_wheel)
        source.Bind(wx.EVT_SET_FOCUS, self.on_focus)
        source.Bind(wx.EVT_KILL_FOCUS, self.on_focus_lost)
        source.Bind(wx.EVT_ENTER_WINDOW, self.on_mouse_enter)
        source.Bind(wx.EVT_LEAVE_WINDOW, self.on_mouse_leave)

    def set_mouse_mode(self, mode=None):
        self.release_mouse()
        mode_log.debug(f"set_mouse_mode: current mode: {self.mouse_mode}")
        if mode is None:
            mode = self.default_mouse_mode_cls
        if not self.is_mouse_mode(mode):
            self.mouse_mode.cleanup()
            mode_log.debug("set_mouse_mode: %s" % mode)
            self.mouse_mode = mode(self)
        else:
            mode_log.debug("mouse mode already %s" % mode)

    def is_mouse_mode(self, mode):
        return self.mouse_mode.__class__ == mode

    def capture_mouse(self):
        if not self.source.HasFocus():
            self.source.SetFocus()
        self.source.CaptureMouse()

    def has_capture(self):
        return self.source.HasCapture()

    def release_mouse(self):
        self.mouse_is_down = False
        self.selection_box_is_being_defined = False
        while self.source.HasCapture():
            self.source.ReleaseMouse()

    def set_cursor(self, mode=None):
        if (self.forced_cursor is not None):
            self.SetCursor(self.forced_cursor)
            #
            return

        if mode is None:
            mode = self.mouse_mode
        c = mode.get_cursor()
        self.SetCursor(c)

    def get_effective_tool_mode(self, event):
        middle_down = False
        alt_down = False
        if (event is not None):
            try:
                alt_down = event.AltDown()
                # print self.is_alt_key_down
            except:
                pass
            try:
                middle_down = event.MiddleIsDown()
            except:
                pass
        if alt_down or middle_down:
            mode = self.default_mouse_mode
        else:
            mode = self.mouse_mode
        return mode

    def on_left_down(self, evt):
        self.capture_mouse()
        if self.is_editing_in_cell and self.mouse_event_in_edit_cell(evt):
            self.is_mousing_while_editing = True
            self.on_left_down_in_edit_cell(evt)
        else:
            self.end_editing()
            mode = self.get_effective_tool_mode(evt)
            mode_log.debug("on_left_down: effective mode=%s" % mode)
            self.forced_cursor = None
            self.selection_box_is_being_defined = False
            self.mouse_down_position = evt.GetPosition()
            self.mouse_move_position = self.mouse_down_position

            mode.process_left_down(evt)
            self.set_cursor(mode)
        evt.Skip()

    def on_motion(self, evt):
        if self.is_mousing_while_editing:
            self.on_motion_in_edit_cell(evt)
        else:
            mode = self.get_effective_tool_mode(evt)
            # mode_log.debug("on_motion: effective mode=%s" % mode)
            if evt.LeftIsDown() and self.has_capture():
                mode.process_mouse_motion_down(evt)
            else:
                mode.process_mouse_motion_up(evt)
            self.set_cursor(mode)
        evt.Skip()

    def on_left_up(self, evt):
        self.stop_scroll_timer()
        if not self.has_capture():
            return
        self.release_mouse()
        if self.is_mousing_while_editing:
            self.is_mousing_while_editing = False
            self.on_left_up_in_edit_cell(evt)
        else:
            mode = self.get_effective_tool_mode(evt)
            mode_log.debug("on_left_up: effective mode=%s" % mode)
            self.forced_cursor = None
            mode.process_left_up(evt)
            self.set_cursor(mode)
        evt.Skip()

    def on_left_dclick(self, evt):
        # self.SetFocus() # why would it not be focused?
        self.end_editing()
        mode = self.get_effective_tool_mode(evt)
        mode_log.debug("on_left_dclick: effective mode=%s" % mode)
        mode.process_left_dclick(evt)
        self.set_cursor(mode)
        evt.Skip()

    def on_popup(self, evt):
        self.end_editing()
        mode = self.get_effective_tool_mode(evt)
        mode_log.debug("on_popup: effective mode=%s" % mode)
        self.forced_cursor = None
        mode.process_popup(evt)
        self.set_cursor(mode)

    def on_mouse_wheel(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode_log.debug("on_mouse_wheel: effective mode=%s" % mode)
        mode.process_mouse_wheel(evt)
        self.set_cursor(mode)

    def on_mouse_enter(self, evt):
        self.set_cursor()
        evt.Skip()

    def on_mouse_leave(self, evt):
        self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
        self.mouse_mode.process_mouse_leave(evt)
        evt.Skip()

    def on_focus(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_focus(evt)

    def on_focus_lost(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_focus_lost(evt)

    ##### mouse processing while in cell edit mode

    def mouse_event_in_edit_cell(self, evt):
        return False

    def on_left_down_in_edit_cell(self, evt):
        pass

    def on_motion_in_edit_cell(self, evt):
        pass

    def on_left_up_in_edit_cell(self, evt):
        pass

    ##### autoscrolling

    def stop_scroll_timer(self):
        pass

    ##### command processor

    def create_mouse_event_flags(self):
        flags = DisplayFlags(self)
        flags.selecting_rows = False
        flags.old_carets = self.caret_handler.calc_state()
        return flags

    def handle_motion_update_status(self, evt, row, col):
        msg = self.get_status_message_at_row_col(row, col)
        if msg:
            self.segment_viewer.show_status_message(msg)

    ##### Default selection handlers

    def handle_select_start(self, evt, row, col, flags=None):
        """ select_handler: interface with set_style_ranges to highlight bytes
        that should be selected (e.g. rect select will need different method
        than regular selection).

        caret_handler: object that implements the CaretHandler API
        """
        if flags is None:
            flags = self.create_mouse_event_flags()
        ch = self.caret_handler
        r, c, inside = self.get_row_col_from_event(evt)
        self.mouse_drag_started = True
        if c < 0 or flags.selecting_rows or not inside:
            c = 0
            selecting_rows = True
            c_end_of_row = self.table.get_items_in_row(r)
        else:
            selecting_rows = False
        mouse_at = (r, c)
        mode_log.debug("handle_select_start: rows=%s, row,col=%d,%d r,c=%d,%d inside=%s" % (selecting_rows, row, col, r, c, inside))
        if evt.ControlDown():
            self.multi_select_mode = True
            self.select_extend_mode = False
        elif evt.ShiftDown():
            self.multi_select_mode = False
            self.select_extend_mode = True
        mode_log.debug(f"start before: {ch.carets}, multi {self.multi_select_mode}, extend {self.select_extend_mode}")
        if self.select_extend_mode:
            caret = ch.current
            if mouse_at < caret.anchor_start:
                self.select_extend_mode = "bottom anchor"
                if not caret.has_selection:
                    caret.anchor_end = mouse_at
                caret.anchor_start = mouse_at
                caret.rc = mouse_at
            else:
                self.select_extend_mode = "top anchor"
                caret.anchor_end = mouse_at
                caret.rc = mouse_at
            caret.anchor_initial_start, caret.anchor_initial_end = caret.anchor_start, caret.anchor_end
        else:
            if selecting_rows:
                ch.move_carets_to(*mouse_at)
                caret = ch.current
                caret.anchor_initial_start = caret.anchor_start = mouse_at
                caret.anchor_initial_end = caret.anchor_end = (r, c_end_of_row)
                flags.keep_selection = True
                mode_log.debug("handle_select_start selecting rows: flags: %s, anchors=%s" % (flags, str((caret.anchor_initial_start, caret.anchor_initial_end))))
            else:
                flags.keep_selection = True
                # initial click when not selecting rows should move the caret,
                # not select the grid square
                if self.multi_select_mode:
                    caret = Caret(*mouse_at)
                    ch.add_caret(caret)
                    mode_log.debug(("adding caret", caret))
                else:
                    ch.move_carets_to(*mouse_at)
                    caret = ch.current
                    mode_log.debug(("forced single caret", caret))
                self.pending_select_awaiting_drag = mouse_at
                mode_log.debug("handle_select_start placing cursor: flags: %s, rc=%s" % (flags, mouse_at))
        self.commit_change(flags)
        mode_log.debug(f"start after: {ch.carets}")

    def handle_select_motion(self, evt, row, col, flags=None):
        if not self.mouse_drag_started:
            # On windows, it's possible to get a motion event before a mouse
            # down event, so need this flag to check
            return
        if flags is None:
            flags = self.create_mouse_event_flags()
        ch = self.caret_handler
        update = flags.refresh_needed
        inside = True ## FIXME
        r = row
        c = col
        mouse_at = (r, c)
        mode_log.debug("handle_select_motion: r=%d c=%d pending: %s, flags: %s" % (r, c, str(self.pending_select_awaiting_drag), flags))
        # mode_log.debug("handle_select_motion: r=%d c=%d index1: %s, index2: %s pending: %s, sel rows: %s anchors: initial=%s current=%s" % (r, c, index1, index2, str(self.pending_select_awaiting_drag), flags.selecting_rows, str((caret.anchor_initial_start_index, caret.anchor_initial_end_index)), str((caret.anchor_start_index, caret.anchor_end_index))))
        mode_log.debug(("motion before:", ch.carets))
        caret = ch.current
        if c < 0 or flags.selecting_rows or not inside:
            selecting_rows = True
            c = 0
        else:
            selecting_rows = False
            if self.pending_select_awaiting_drag is not None:
                # only start selection if the cursor is over a different cell
                # than the mouse down event
                if mouse_at == self.pending_select_awaiting_drag:
                    return

                # We have an actual drag so we can begin the selection
                caret.anchor_initial_start = caret.anchor_initial_end = self.pending_select_awaiting_drag
                caret.anchor_start = caret.anchor_end = self.pending_select_awaiting_drag
                self.pending_select_awaiting_drag = None
                update = True

        if self.select_extend_mode == "top anchor":
            if mouse_at < caret.anchor_initial_start:
                caret.rc = mouse_at
                caret.anchor_start = mouse_at
                caret.anchor_end = caret.anchor_initial_start
                update = True
            else:
                caret.rc = mouse_at
                caret.anchor_start = caret.anchor_initial_start
                caret.anchor_end = mouse_at
                update = True
        elif self.select_extend_mode == "bottom anchor":
            if mouse_at < caret.anchor_initial_end:
                caret.rc = mouse_at
                caret.anchor_start = mouse_at
                caret.anchor_end = caret.anchor_initial_end
                update = True
            else:
                caret.rc = mouse_at
                caret.anchor_start = caret.anchor_initial_end
                caret.anchor_end = mouse_at
                update = True
        else:
            if mouse_at >= caret.anchor_initial_end:
                caret.rc = mouse_at
                if mouse_at != caret.anchor_end:
                    caret.anchor_start = caret.anchor_initial_start
                    caret.anchor_end = mouse_at
                    update = True
            elif mouse_at <= caret.anchor_initial_start:
                caret.rc = mouse_at
                if mouse_at != caret.anchor_start:
                    caret.anchor_start = mouse_at
                    caret.anchor_end = caret.anchor_initial_end
                    update = True

        if update:
            self.commit_change(flags)
        mode_log.debug("handle_select_motion: update: %s, flags: %s, anchors: initial=%s current=%s" % (update, flags, str((caret.anchor_initial_start, caret.anchor_initial_end)), str((caret.anchor_start, caret.anchor_end))))
        # mode_log.debug(f"motion after: {ch.carets}")

    def handle_select_end(self, evt, row, col, flags=None):
        if flags is None:
            flags = self.create_mouse_event_flags()
        self.mouse_drag_started = False
        self.select_extend_mode = False
        self.multi_select_mode = False
        self.caret_handler.collapse_overlapping()
        if wx.Platform == "__WXMSW__":
            # FIXME: MSW doesn't seem to refresh after a mouse release
            # outside of the window, so force it here to fill in the remaining
            # bits of the selection
            mode_log.debug("Extra refresh on handle_select_end for windows")
            self.refresh_view()
        self.commit_change(flags)
        mode_log.debug(("end after:", self.caret_handler.carets))

    def commit_change(self, flags):
        mode_log.debug(("commit before:", self.caret_handler.carets))
        self.mouse_mode.refresh_ranges(self.caret_handler)
        self.send_caret_event(flags)
        # self.caret_handler.sync_caret_event = flags
        # self.caret_handler.ensure_visible_event = flags
        # self.caret_handler.refresh_event = flags
        mode_log.debug(("commit after:", self.caret_handler.carets))

    def get_start_end_index_of_row(self, row):
        raise NotImplementedError

    def get_status_message_at_row_col(self, row, col):
        raise NotImplementedError

    def highlight_selected_ranges(self, caret_handler):
        ch = self.caret_handler
        ch.document.change_count += 1
        s = ch.segment
        s.clear_style_bits(selected=True)
        self.segment_viewer.highlight_selected_ranges_in_segment(ch.carets.selected_ranges, s)
        ch.calc_dependent_action_enabled_flags()

    ##### editing

    def handle_char_ordinary(self, evt):
        c = evt.GetKeyCode()
        print("ordinary char: {c}")
        if not self.is_editing_in_cell:
            print("handle_char_ordinary: not editing in cell")
            if self.verify_keycode_can_start_edit(c):
                self.start_editing(evt)
            else:
                evt.Skip()
        else:
            print("handle_char_ordinary: editing in cell")
            self.edit_source.EmulateKeyPress(evt)

    def verify_keycode_can_start_edit(self, c):
        return True

    def mouse_event_in_edit_cell(self, evt):
        r, c, _ = self.get_row_col_from_event(evt)
        index, _ = self.table.get_index_range(r, c)
        print(("mouse edit cell check: r,c=%d,%d, index=%d" % (r, c, index)))
        return self.caret_handler.is_index_of_caret(index)

    def on_left_down_in_edit_cell(self, evt):
        pass

    def on_motion_in_edit_cell(self, evt):
        pass

    def on_left_up_in_edit_cell(self, evt):
        pass

    def start_editing(self, evt):
        self.is_editing_in_cell = True
        self.edit_source = self.create_hidden_text_ctrl()
        self.edit_source.SetFocus()
        if self.use_first_char_when_starting_edit():
            print(("EmulateKeyPress: %s" % evt.GetKeyCode()))
            self.edit_source.EmulateKeyPress(evt)

    def use_first_char_when_starting_edit(self):
        return True

    def accept_edit(self, autoadvance=False):
        val = self.edit_source.get_processed_value()
        self.end_editing()
        print(("changing to %s" % val))
        self.process_edit(val)

    def process_edit(self, val):
        ranges = []
        # for c in self.caret_handler.carets:
        #     ranges.append((c.index, c.index + 1))
        ranges = self.get_selected_ranges_including_carets()
        cmd = SetRangeValueCommand(self.segment_viewer.segment, ranges, val, advance=True)
        self.segment_viewer.editor.process_command(cmd)

    def end_editing(self):
        if self.is_editing_in_cell:
            self.edit_source.Destroy()
            self.edit_source = None
            self.is_editing_in_cell = False
            self.SetFocus()

    def create_hidden_text_ctrl(self):
        c = GridCellTextCtrl(self, -1, 0, pos=(600,100), size=(400,24))
        return c



# peppy Copyright (c) 2006-2010 Rob McMullen
# Licenced under the GPLv2; see http://peppy.flipturn.org for more info
"""Utility functions for interacting with emacs

These text utilities have no dependencies on any other part of peppy, and
therefore may be used independently of peppy.

The Emacs manual has a description of how U{file local
variables<http://www.gnu.org/software/emacs/manual/html_node/emacs/Specifying-File-Variables.html#Specifying-File-Variables>} are stored.
"""
import re


def parseModeline(line):
    """Determine if line specifies a major mode.
    
    Parse a potential emacs major mode specifier line into the
    mode and the optional variables.  The mode may appears as any
    of::

      -*-C++-*-
      -*- mode: Python; -*-
      -*- mode: Ksh; var1:value1; var3:value9; -*-

    @param header: first x bytes of the file to be loaded
    @return: two-tuple of the mode and a dict of the name/value pairs.
    @rtype: tuple
    """
    match=re.search(r'-\*\-\s*(mode:\s*(.+?)|(.+?))(\s*;\s*(.+?))?\s*-\*-',line)
    vars={}
    if match:
        varstring=match.group(5)
        if varstring:
            try:
                for nameval in varstring.split(';'):
                    s=nameval.strip()
                    if s:
                        name,val=s.split(':')
                        vars[name.strip()]=val.strip()
            except:
                pass
        if match.group(2):
            return (match.group(2),vars)
        elif match.group(3):
            return (match.group(3),vars)
    return None, vars


def applyEmacsFileLocalSettings(stc):
    """Scan the stc for file local settings.
    
    Emacs uses U{file local
    variables<http://www.gnu.org/software/emacs/manual/html_node/emacs/Specifying-    File-Variables.html#Specifying-File-Variables>}
    to customize major modes on a per-file basis.  This method parses both
    the modeline specifier (the second or first line depending on whether the
    file is a shell script or not) and the local variables list at the end of
    the file.
    
    Emacs modes can have many local variables, so it's left to the major mode
    itself to parse most of them.  There are, however, some U{standard buffer
    local variables<http://www.gnu.org/software/emacs/elisp/html_node/Standard-Buffer_002dLocal-Variables.html>}
    that emacs does recognize.
    
    @param stc: styled text control that will be used to apply settings
    
    @return: list of settings affected.  Settings changed are reported as the
    name of the wx.stc method used for each setting, with the 'Set' removed.
    If the tab width is changed, the setting is reported as 'TabWidth'.
    """
    modeline = stc.GetLine(0)
    if modeline.startswith("#!") and stc.GetLineCount() > 1:
        modeline = stc.GetLine(1)
    mode, vars = parseModeline(modeline)

    settings_changed = []

    # check for integers.  int_mapping takes the emacs string to the name of
    # the stc getter/setter function.
    int_mapping = {'fill-column': 'EdgeColumn',
                   'tab-width': 'TabWidth',
                   }
    for name, setting in list(int_mapping.items()):
        if name in vars:
            # Construct the name of the stc setting function and set the value
            func = getattr(stc, "Set%s" % setting)
            func(int(vars[name]))
            settings_changed.append(setting)

    # check for booleans -- emacs uses 'nil' for false, everything else for true
    bool_mapping = {'use-tabs': 'UseTabs',
                    }
    for name, setting in list(bool_mapping.items()):
        if name in vars:
            func = getattr(stc, "Set%s" % setting)
            func(vars[name] != 'nil')
            settings_changed.append(setting)

    # check for more complicated settings
    if 'cursor-type' in vars:
        text = vars['cursor-type']
        match = re.search(r'\(\s*bar\s*\.\s*([0-9]+)', text)
        if match:
            width = int(match.group(1))
            stc.SetCaretWidth(width)
            settings_changed.append('CaretWidth')

    return settings_changed, vars



import os

import wx

from ..editor import SawxEditor
from ..filesystem import fsopen as open
from ..keybindings import KeyBindingControlMixin
from ..preferences import SawxEditorPreferences

import logging
log = logging.getLogger(__name__)


class TextEditorControl(KeyBindingControlMixin, wx.TextCtrl):
    # FIXME: is this second level of keyboard processing necessary now that the
    # on_char_hook event raises a ProcessKeystrokeNormally error, allowing the
    # control to process that keytroke?
    keybinding_desc = {
        "prev_line": "Up",
        "next_line": "Down",
        "prev_char": "Left",
        "next_char": "Right",
        "delete_selection": "Delete",
    }

    def __init__(self, *args, **kwargs):
        wx.TextCtrl.__init__(self, *args, **kwargs)
        KeyBindingControlMixin.__init__(self)

    def prev_line(self, evt):
        print("prev_line")

    def next_line(self, evt):
        print("next_line")

    def prev_char(self, evt):
        print("prev_char")

    def next_char(self, evt):
        print("next_char")

    def delete_selection(self, evt):
        print("delete_selection")


class TextEditorPreferences(SawxEditorPreferences):
    pass


class TextEditor(SawxEditor):
    editor_id = "text_editor"

    ui_name = "Text Editor"

    preferences_module = "sawx.editors.text_editor"

    @property
    def is_dirty(self):
        return self.control.IsModified()

    @property
    def can_copy(self):
        return self.control.CanCopy()

    @property
    def can_undo(self):
        return self.control.CanUndo()

    @property
    def can_redo(self):
        return self.control.CanRedo()

    def create_control(self, parent):
        # return TextEditorControl(parent, -1, style=wx.TE_MULTILINE)
        return wx.TextCtrl(parent, -1, style=wx.TE_MULTILINE)

    def create_event_bindings(self):
        self.control.Bind(wx.EVT_CONTEXT_MENU, self.on_popup)

    def show(self, args=None):
        self.control.SetValue(self.document.raw_data)
        if args is not None:
            size = int(args.get('size', -1))
            if size > 0:
                font = self.control.GetFont()
                font.SetPointSize(size)
                self.control.SetFont(font)
        self.control.DiscardEdits()

    def save(self):
        self.document.raw_data = self.control.GetValue()
        super().save()
        self.control.DiscardEdits()

    @classmethod
    def can_edit_document_exact(cls, document):
        return document.mime == "text/plain"

    @classmethod
    def can_edit_document_generic(cls, document):
        return document.mime.startswith("text/")

    def on_popup(self, evt):
        popup_menu_desc = [
            "undo",
            "redo",
            None,
            "copy",
            "cut",
            "paste",
        ]
        self.show_popup(popup_menu_desc)

    #### selection

    def select_all(self):
        self.control.SelectAll()

    def select_none(self):
        self.control.SelectNone()

    def select_invert(self):
        start, end = self.control.GetSelection()
        if start == end:
            self.control.SelectAll()
        elif start == 0 and end == self.control.GetLastPosition():
            self.control.SelectNone()
        else:
            # need to implement multi-select here
            pass

    #### copy/paste stuff

    supported_clipboard_handlers = [
        (wx.TextDataObject(), "paste_text_control"),
    ]

    def delete_selection_from(self, focused):
        start, end = focused.GetSelection()
        focused.Remove(start, end)


class DebugTextEditor(TextEditor):
    editor_id = "debug"

    ui_name = "Debug Text Editor"

    menubar_desc = [
        ["File",
            ["New",
                "new_file",
            ],
            "new_window",
            None,
            "open_file",
            ["Open Recent",
                "open_recent",
            ],
            None,
            "save_file",
            "save_as",
            None,
            "quit",
        ],
        ["Edit",
            "undo",
            "redo",
            None,
            "copy",
            "cut",
            "paste",
            ["Paste Special",
                "paste_as_text",
                "paste_as_hex",
            ],
            None,
            "select_all",
            "select_none",
            "select_invert",
            None,
            "prefs",
        ],
        ["Debug",
            None,
            None,
            None,
            "debug_text_counting",
            None,
            None,
            None,
            "debug_text_last_digit",
            None,
            "debug_text_size",
        ],
        ["Dynamic",
            "debug_text_last_digit_dyn",
        ],
        ["Generated",
            "debug_gen_sub_menu()",
            None,
            "debug_generated{1}",
            "debug_generated{34}",
            "debug_generated{whatever}",
        ],
        ["Help",
            "about",
        ],
    ]

    toolbar_desc = [
    "new_file", "open_file", "save_file", None, "undo", "redo", None, "copy", "cut", "paste", "paste_as_text", "paste_as_hex",
    ]

    keybinding_desc = {
        "new_file": "Ctrl+N",
        "open_file": "Ctrl+O",
        "save_file" : "Ctrl+S",
        "save_as" : "Shift+Ctrl+S",
        "cut": "Ctrl+X",
        "copy": "Ctrl+C",
        "paste": "Ctrl+V",
        "prev_line": "Up",
        "next_line": "Down",
        "prev_char": "Left",
        "next_char": "Right",
        "delete_selection": "Delete",
    }

    @property
    def tab_name(self):
        return "DEBUG " + super().tab_name

    # won't automatically match anything; must force this editor with the -t
    # command line flag
    @classmethod
    def can_edit_file_exact(cls, file_metadata):
        return False

    @classmethod
    def can_edit_file_generic(cls, file_metadata):
        return file_metadata['mime'].startswith("text/")

    def debug_gen_sub_menu(self):
        return [
            ["Sub Menu 1", "about", ["Sub Sub Menu 1!", "about"]],
            ["Sub Menu 2", "about", "about",],
            ["Sub Menu 3", "about", "about", "about", "debug_gen_sub_sub_menu()"],
            ["Sub Menu 4", "about", "about", "about", "debug_gen_sub_sub_menu(5)"],
        ]

    def debug_gen_sub_sub_menu(self, count="2"):
        menu = ["Sub Sub Menu 2!"]
        count = int(count)
        for i in range(count):
            menu.append("about")
        return [menu]



from ..action import SawxListAction
from .. import errors
from ..persistence import get_json_data, save_json_data

class open_recent(SawxListAction):
    prefix = "open_recent_"

    max_entries = 20

    canonical_list = None

    @classmethod
    def reload(cls):
        data = get_json_data("recent_files", [])
        cls.canonical_list = data

    @classmethod
    def append(cls, uri):
        if cls.canonical_list is None:
            cls.reload()
        try:
            index = cls.canonical_list.index(uri)
            cls.canonical_list.pop(index)
        except ValueError:
            pass
        cls.canonical_list[0:0] = [uri]
        if len(cls.canonical_list) > cls.max_entries:
            cls.canonical_list = cls.canonical_list[:cls.max_entries]
        save_json_data("recent_files", cls.canonical_list)

    def calc_list_items(self):
        if self.canonical_list is None:
            self.reload()
        return self.canonical_list[:]

    def get_path(self, action_key):
        return self.current_list[self.get_index(action_key)]

    def sync_menu_item_from_editor(self, action_key, menu_item):
        if self.current_list != self.canonical_list:
            raise errors.RecreateDynamicMenuBar
        SawxListAction.sync_menu_item_from_editor(self, action_key, menu_item)

    def perform(self, action_key):
        path = self.get_path(action_key)
        self.editor.frame.load_file(path, self.editor)



#!/usr/bin/env python

# Standard library imports.
import sys
import logging


last_trace_was_system_call = False
trace_after_funcname = None

def trace_calls(frame, event, arg):
    global last_trace_was_system_call, trace_after_funcname

    if event != 'call':
        return
    co = frame.f_code
    func_name = co.co_name
    if func_name == 'write':
        # Ignore write() calls from print statements
        return
    if trace_after_funcname is not None:
        if func_name == trace_after_funcname:
            trace_after_funcname = None
        else:
            # skip anything until it hits the trace_after function
            return
    func_line_no = frame.f_lineno
    func_filename = co.co_filename
    caller = frame.f_back
    caller_line_no = caller.f_lineno
    caller_filename = caller.f_code.co_filename
    if "/python3" in caller_filename or "/logging/" in func_filename or "/wx/core.py" in func_filename or "/sre_" in caller_filename or "/logging/" in caller_filename:
        if not last_trace_was_system_call:
            print('  <system calls>')
            last_trace_was_system_call = True
        return
    last_trace_was_system_call = False
    print('%s:%s -> %s %s:%s' % (caller_filename, caller_line_no, func_name, func_filename, func_line_no))
    return

def create_global_functions():
    def what_called_me():
        import traceback
        stack = traceback.extract_stack()
        count = len(stack) - 2
        for i, item in enumerate(stack[:-1]):
            print(("#%d %s in %s at %s:%d" % (count - i, item[3], item[2], item[0], item[1])))
    import builtins
    builtins.what_called_me = what_called_me

create_global_functions()

def main(argv):
    """ Run the application.
    """
    logging.basicConfig(level=logging.WARNING)

    if "--trace" in argv:
        i = argv.index("--trace")
        argv.pop(i)
        sys.settrace(trace_calls)

    if "--trace-after" in argv:
        global trace_after_funcname
        i = argv.index("--trace-after")
        argv.pop(i)
        funcname = argv.pop(i)
        trace_after_funcname = funcname
        sys.settrace(trace_calls)

    from sawx.startup import run
    from sawx.application import SawxApp
    from sawx._version import __version__
    SawxApp.app_version = __version__

    run(SawxApp)

    logging.shutdown()


if __name__ == '__main__':
    import sys
    from sawx.startup import setup_frozen_logging
    
    setup_frozen_logging()
    main(sys.argv)



"""General purpose event handling routines.

From public domain code by Marcus von Appen, described at:
https://python-utilities.readthedocs.io/en/latest/events.html
"""
import weakref

__all__ = ["EventHandler"]

import logging
log = logging.getLogger(__name__)


class Event(list):
    def __init__(self, sender, *args, **kwargs):
        self._sender = sender
        self[:] = args
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def sender(self):
        return self._sender


class EventHandler:
    """A simple event handling class, which manages callbacks to be
    executed.
    """
    event_class = Event

    def __init__(self, sender, event_class=None, debug=False):
        self.callbacks = []
        self.sender = sender
        if not debug:
            debug = log.getEffectiveLevel() <= logging.DEBUG
        self.debug = debug

    def __call__(self, *args, **kwargs):
        """Executes all callbacks.

        Executes all connected callbacks in the order of addition,
        passing the sender of the EventHandler as first argument and the
        optional args as second, third, ... argument to them.
        """
        results = []
        remove_indexes = []
        evt = self.event_class(self.sender, *args, **kwargs)
        for i, ref in enumerate(self.callbacks):
            callback = ref()
            if callback is None:
                if self.debug:
                    print(f"EventHandler {self.sender}: callback {ref} deleted; removing from callback list")
                remove_indexes.append(i)
            else:
                if self.debug:
                    print(f"EventHandler {self.sender}: calling {callback}, args={args}")
                result = callback(evt)
                results.append(result)
        try:
            while True:
                i = remove_indexes.pop()
                del self.callbacks[i]
        except IndexError:
            pass
        return results

    def __iadd__(self, callback):
        """Adds a callback to the EventHandler."""
        self.add(callback)
        return self

    def __isub__(self, callback):
        """Removes a callback from the EventHandler."""
        self.remove(callback)
        return self

    def __len__(self):
        """Gets the amount of callbacks connected to the EventHandler."""
        return len(self.callbacks)

    def __getitem__(self, index):
        return self.callbacks[index]

    def __setitem__(self, index, value):
        self.callbacks[index] = value

    def __delitem__(self, index):
        del self.callbacks[index]

    def add(self, callback):
        """Adds a callback to the EventHandler."""
        if not callable(callback):
            raise TypeError("callback must be callable")
        if hasattr(callback, "__self__"):
            # bound methods are ephemeral objects so weak refs can't keep them
            ref = weakref.WeakMethod(callback)
            if self.debug:
                print(f"EventHandler {self.sender}: {callback} is a bound method; using WeakMethod")
        else:
            ref = weakref.ref(callback)
            if self.debug:
                print(f"EventHandler {self.sender}: {callback} using a normal weak reference")
        if self.debug:
            print(f"EventHandler {self.sender}: adding {callback}, ref={ref}")
        self.callbacks.append(ref)

    def remove(self, callback):
        """Removes a callback from the EventHandler."""
        if self.debug:
            print(f"EventHandler {self.sender}: removing {callback}")
        for i, ref in enumerate(self.callbacks):
            if ref() == callback:
                if self.debug:
                    print(f"EventHandler {self.sender}: found {callback} at {i}")
                del self.callbacks[i]
                break
        else:
            raise ValueError(f"EventHandler {self.sender}: callback {callback} not found")
